# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'tikka/slots/pyqt/resources/gui/widgets/currency.ui'
#
# Created by: PyQt5 UI code generator 5.15.9
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_currencyWidget(object):
    def setupUi(self, currencyWidget):
        currencyWidget.setObjectName("currencyWidget")
        currencyWidget.resize(383, 169)
        self.gridLayout = QtWidgets.QGridLayout(currencyWidget)
        self.gridLayout.setObjectName("gridLayout")
        self.nameValueLabel = QtWidgets.QLabel(currencyWidget)
        self.nameValueLabel.setText("")
        self.nameValueLabel.setObjectName("nameValueLabel")
        self.gridLayout.addWidget(self.nameValueLabel, 1, 1, 1, 1)
        self.monetaryMassLabel = QtWidgets.QLabel(currencyWidget)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.monetaryMassLabel.setFont(font)
        self.monetaryMassLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.monetaryMassLabel.setObjectName("monetaryMassLabel")
        self.gridLayout.addWidget(self.monetaryMassLabel, 4, 0, 1, 1)
        self.universalDividendValueLabel = QtWidgets.QLabel(currencyWidget)
        self.universalDividendValueLabel.setText("")
        self.universalDividendValueLabel.setObjectName("universalDividendValueLabel")
        self.gridLayout.addWidget(self.universalDividendValueLabel, 3, 1, 1, 1)
        self.universalDividendLabel = QtWidgets.QLabel(currencyWidget)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.universalDividendLabel.setFont(font)
        self.universalDividendLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.universalDividendLabel.setObjectName("universalDividendLabel")
        self.gridLayout.addWidget(self.universalDividendLabel, 3, 0, 1, 1)
        self.membersValueLabel = QtWidgets.QLabel(currencyWidget)
        self.membersValueLabel.setText("")
        self.membersValueLabel.setObjectName("membersValueLabel")
        self.gridLayout.addWidget(self.membersValueLabel, 5, 1, 1, 1)
        self.membersLabel = QtWidgets.QLabel(currencyWidget)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.membersLabel.setFont(font)
        self.membersLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.membersLabel.setObjectName("membersLabel")
        self.gridLayout.addWidget(self.membersLabel, 5, 0, 1, 1)
        self.monetaryMassValueLabel = QtWidgets.QLabel(currencyWidget)
        self.monetaryMassValueLabel.setText("")
        self.monetaryMassValueLabel.setObjectName("monetaryMassValueLabel")
        self.gridLayout.addWidget(self.monetaryMassValueLabel, 4, 1, 1, 1)
        self.nameLabel = QtWidgets.QLabel(currencyWidget)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.nameLabel.setFont(font)
        self.nameLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.nameLabel.setObjectName("nameLabel")
        self.gridLayout.addWidget(self.nameLabel, 1, 0, 1, 1)
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.gridLayout.addItem(spacerItem, 6, 0, 1, 1)
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.gridLayout.addItem(spacerItem1, 1, 2, 1, 1)
        self.refreshButton = QtWidgets.QPushButton(currencyWidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.refreshButton.sizePolicy().hasHeightForWidth())
        self.refreshButton.setSizePolicy(sizePolicy)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/icons/refresh_icon"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.refreshButton.setIcon(icon)
        self.refreshButton.setObjectName("refreshButton")
        self.gridLayout.addWidget(self.refreshButton, 0, 2, 1, 1)
        self.symbolLabel = QtWidgets.QLabel(currencyWidget)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.symbolLabel.setFont(font)
        self.symbolLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.symbolLabel.setObjectName("symbolLabel")
        self.gridLayout.addWidget(self.symbolLabel, 2, 0, 1, 1)
        self.symbolValueLabel = QtWidgets.QLabel(currencyWidget)
        self.symbolValueLabel.setText("")
        self.symbolValueLabel.setObjectName("symbolValueLabel")
        self.gridLayout.addWidget(self.symbolValueLabel, 2, 1, 1, 1)

        self.retranslateUi(currencyWidget)
        QtCore.QMetaObject.connectSlotsByName(currencyWidget)

    def retranslateUi(self, currencyWidget):
        #  _translate = QtCore.QCoreApplication.translate
        currencyWidget.setWindowTitle(_("Currency parameters"))
        self.monetaryMassLabel.setText(_("Monetary Mass"))
        self.universalDividendLabel.setText(_("Universal Dividend"))
        self.membersLabel.setText(_("Members"))
        self.nameLabel.setText(_("Name"))
        self.refreshButton.setText(_("Refresh"))
        self.symbolLabel.setText(_("Symbol"))
import tikka.slots.pyqt.resources.icons.index_rc



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from dataclasses import dataclass
from typing import Optional
from uuid import UUID


@dataclass
class Category:
    id: UUID
    name: str
    expanded: bool = True
    parent_id: Optional[UUID] = None

    def __eq__(self, other):
        """
        Test equality on id

        :param other: Category instance
        :return:
        """
        if not isinstance(other, self.__class__):
            return False
        return other.id == self.id

    def __hash__(self):
        return hash(self.id)



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import sys
from typing import Optional

from PyQt5.QtCore import QMutex, QPoint
from PyQt5.QtWidgets import QApplication, QMainWindow, QWidget

from tikka.domains.application import Application
from tikka.domains.entities.constants import DATA_PATH
from tikka.domains.entities.events import CurrencyEvent, NodesEvent
from tikka.slots.pyqt.models.nodes import NodesTableModel
from tikka.slots.pyqt.resources.gui.widgets.nodes_rc import Ui_NodesWidget
from tikka.slots.pyqt.widgets.node_menu import NodePopupMenu
from tikka.slots.pyqt.windows.node_add import NodeAddWindow


class NodesWidget(QWidget, Ui_NodesWidget):
    """
    NodesWidget class
    """

    def __init__(
        self,
        application: Application,
        mutex: QMutex,
        parent: Optional[QWidget] = None,
    ) -> None:
        """
        Init NodesWidget instance

        :param application: Application instance
        :param mutex: QMutex instance
        :param parent: MainWindow instance
        """
        super().__init__(parent=parent)
        self.setupUi(self)

        self.application = application
        self.mutex = mutex
        self._ = self.application.translator.gettext

        self.nodes_table_model = NodesTableModel(self.application)
        self.nodesTableView.setModel(self.nodes_table_model)
        self.nodesTableView.resizeColumnsToContents()
        self.nodesTableView.resizeRowsToContents()

        self.nodesTableView.customContextMenuRequested.connect(self.on_context_menu)

        # events
        self.addServerButton.clicked.connect(self._on_add_server_button_clicked)

        # subscribe to application events
        self.application.event_dispatcher.add_event_listener(
            CurrencyEvent.EVENT_TYPE_CHANGED, self._on_currency_event
        )
        self.application.event_dispatcher.add_event_listener(
            NodesEvent.EVENT_TYPE_LIST_CHANGED, self._on_node_list_changed_event
        )

    def _on_add_server_button_clicked(self):
        """
        Trigger when user click on add server button

        :return:
        """
        NodeAddWindow(self.application, self).exec_()

    def _on_currency_event(self, _):
        """
        When a currency event is triggered

        :param _: CurrencyEvent instance
        :return:
        """
        # update model
        self.nodesTableView.model().init_data()
        # resize view
        self.nodesTableView.resizeColumnsToContents()
        self.nodesTableView.resizeRowsToContents()

    def _on_node_list_changed_event(self, _):
        """
        When the node list has changed

        :param _: NodesEvent instance
        :return:
        """
        # update model
        self.nodesTableView.model().init_data()
        # resize view
        self.nodesTableView.resizeColumnsToContents()
        self.nodesTableView.resizeRowsToContents()

    def on_context_menu(self, position: QPoint):
        """
        When right button on table view

        :param position: QPoint instance
        :return:
        """
        index = self.nodesTableView.indexAt(position)
        if index.isValid():
            # get selected node
            row = index.row()
            node = self.nodes_table_model.nodes[row]
            # display popup menu at click position
            NodePopupMenu(self.application, node).exec_(
                self.nodesTableView.mapToGlobal(position)
            )


if __name__ == "__main__":
    qapp = QApplication(sys.argv)

    application_ = Application(DATA_PATH)

    main_window = QMainWindow()
    main_window.show()

    main_window.setCentralWidget(NodesWidget(application_, QMutex(), main_window))

    sys.exit(qapp.exec_())



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import sys
from pathlib import Path

from PyQt5.QtCore import QCoreApplication, QLibraryInfo, QLocale, QTranslator
from PyQt5.QtGui import QIcon
from PyQt5.QtWidgets import QApplication

from tikka.domains.application import Application
from tikka.domains.config import Config
from tikka.interfaces.main import MainApplicationInterface
from tikka.slots.pyqt.windows.main import MainWindow


class PyQtMainApplication(MainApplicationInterface):
    """
    PyQtMainApplication class
    """

    def __init__(
        self, application: Application
    ):  # pylint: disable=super-init-not-called
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            MainApplicationInterface.__init__.__doc__
        )
        qapp = QApplication(sys.argv)

        # qt builtin locales for filedialog or buttonbox...
        QLocale.setDefault(QLocale(application.config.get(Config.LANGUAGE_KEY)))
        translator = QTranslator(qapp)
        filename = "qtbase_" + application.config.get(Config.LANGUAGE_KEY).split("_")[0]
        result = translator.load(
            filename, QLibraryInfo.location(QLibraryInfo.TranslationsPath)
        )
        if result is False:
            raise FileNotFoundError(
                Path(QLibraryInfo.location(QLibraryInfo.TranslationsPath)).joinpath(
                    filename
                )
            )
        QCoreApplication.installTranslator(translator)

        # main window
        qapp.setWindowIcon(QIcon(":/icons/logo"))
        main_window = MainWindow(application)
        main_window.show()

        # open welcome window if no accounts
        if application.accounts.count() == 0:
            main_window.open_welcome_window()

        sys.exit(qapp.exec_())



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from datetime import datetime, timedelta
from typing import Optional

from tikka.domains.currencies import Currencies
from tikka.domains.nodes import Nodes


class BlockTime:
    def __init__(self, currencies: Currencies, nodes: Nodes):
        """
        Initialize BlockTime domain instance

        :param currencies: Currencies domain instance
        :param nodes: Nodes domain instance
        """
        self.currencies = currencies
        self.nodes = nodes

    def get_datetime_from_block(self, block_number: int) -> Optional[datetime]:
        """
        Return a datetime object from a block number

        :param block_number: Block number
        :return:
        """
        # network access to update current block number, can be slow...
        self.nodes.network_fetch_current_node()

        current_time = datetime.now()
        node = self.nodes.get(self.nodes.get_current_url())
        if node is None:
            return None
        currency = self.currencies.get_current()
        current_block_number = node.block
        if current_block_number is None:
            return None
        block_diff = block_number - current_block_number
        if block_diff < 0:
            block_time = current_time - timedelta(
                milliseconds=abs(block_diff) * currency.block_duration
            )
        else:
            block_time = current_time + timedelta(
                milliseconds=abs(block_diff) * currency.block_duration
            )

        return block_time



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from dataclasses import dataclass
from typing import Optional, TypeVar
from uuid import UUID

AccountType = TypeVar("AccountType", bound="Account")


@dataclass
class Account:

    address: str
    name: Optional[str] = None
    crypto_type: Optional[int] = None
    balance: Optional[int] = None
    identity_index: Optional[int] = None
    path: Optional[str] = None
    root: Optional[str] = None
    file_import: bool = False
    category_id: Optional[UUID] = None

    def __str__(self):
        """
        Return string representation

        :return:
        """
        if self.name:
            return f"{self.address} - {self.name}"
        return f"{self.address}"

    def __eq__(self, other):
        """
        Test equality on address

        :param other: Account instance
        :return:
        """
        if not isinstance(other, self.__class__):
            return False
        return other.address == self.address

    def __hash__(self):
        return hash(self.address)



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import List, Optional

from tikka.adapters.repository.sqlite3 import Sqlite3RepositoryInterface
from tikka.domains.entities.node import Node
from tikka.interfaces.adapters.repository.nodes import NodesRepositoryInterface

TABLE_NAME = "nodes"

SQL_COLUMNS = {
    NodesRepositoryInterface.COLUMN_URL: "url",
    NodesRepositoryInterface.COLUMN_PEER_ID: "peer_id",
    NodesRepositoryInterface.COLUMN_BLOCK: "block",
    NodesRepositoryInterface.COLUMN_SOFTWARE: "software",
    NodesRepositoryInterface.COLUMN_SOFTWARE_VERSION: "software_version",
    NodesRepositoryInterface.COLUMN_SESSION_KEYS: "session_keys",
}

DEFAULT_LIST_OFFSET = 0
DEFAULT_LIST_LIMIT = 1000


class Sqlite3NodesRepository(NodesRepositoryInterface, Sqlite3RepositoryInterface):
    """
    Sqlite3NodesRepository class
    """

    def list(
        self,
        offset: int = DEFAULT_LIST_OFFSET,
        limit: int = DEFAULT_LIST_LIMIT,
    ) -> List[Node]:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            NodesRepositoryInterface.list.__doc__
        )

        sql = f"SELECT * FROM {TABLE_NAME} LIMIT {limit} OFFSET {offset}"
        result_set = self.client.select(sql)
        list_ = []
        for row in result_set:
            list_.append(Node(*row))

        return list_

    def add(self, node: Node) -> None:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            NodesRepositoryInterface.add.__doc__
        )
        # insert only non protected fields
        self.client.insert(
            TABLE_NAME,
            **get_fields_from_node(node),
        )

    def get(self, url: str) -> Optional[Node]:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            NodesRepositoryInterface.list.__doc__
        )

        row = self.client.select_one(f"SELECT * FROM {TABLE_NAME} WHERE url=?", (url,))
        if row is None:
            return None

        return Node(*row)

    def update(self, node: Node) -> None:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            NodesRepositoryInterface.update.__doc__
        )

        # update only non hidden fields
        self.client.update(
            TABLE_NAME,
            f"url='{node.url}'",
            **get_fields_from_node(node),
        )

    def delete(self, url: str) -> None:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            NodesRepositoryInterface.delete.__doc__
        )

        self.client.delete(TABLE_NAME, url=url)

    def delete_all(self) -> None:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            NodesRepositoryInterface.delete.__doc__
        )

        self.client.clear(TABLE_NAME)

    def count(self) -> int:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            NodesRepositoryInterface.delete.__doc__
        )
        row = self.client.select_one(f"SELECT count(url) FROM {TABLE_NAME}")
        if row is None:
            return 0

        return row[0]

    def get_urls(self) -> List[str]:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            NodesRepositoryInterface.get_urls.__doc__
        )
        sql = f"SELECT url FROM {TABLE_NAME}"
        result_set = self.client.select(sql)
        list_ = []
        for row in result_set:
            list_.append(row[0])

        return list_


def get_fields_from_node(node: Node) -> dict:
    """
    Return a dict of supported fields with normalized value

    :param node: Node instance
    :return:
    """
    fields = {}
    for (key, value) in node.__dict__.items():
        if key.startswith("_"):
            continue
        fields[key] = value

    return fields


def get_node_from_row(row: tuple) -> Node:
    """
    Return a Node instance from a result set row

    :param row: Result set row
    :return:
    """
    values = list(row)

    return Node(*values)



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from typing import List

from tikka.adapters.repository.sqlite3 import Sqlite3RepositoryInterface
from tikka.domains.entities.tab import Tab
from tikka.interfaces.adapters.repository.tabs import TabRepositoryInterface

TABLE_NAME = "tabs"


class Sqlite3TabRepository(TabRepositoryInterface, Sqlite3RepositoryInterface):
    """
    Sqlite3TabRepository class
    """

    def add(self, tab: Tab) -> None:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            TabRepositoryInterface.add.__doc__
        )

        self.client.insert(
            TABLE_NAME,
            **{
                key: value
                for (key, value) in tab.__dict__.items()
                if not key.startswith("_")
            },
        )

    def delete(self, id: str) -> None:  # pylint: disable=redefined-builtin
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            TabRepositoryInterface.delete.__doc__
        )

        self.client.delete(TABLE_NAME, id=id)

    def delete_all(self) -> None:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            TabRepositoryInterface.delete_all.__doc__
        )
        # remove all entries
        self.client.clear(TABLE_NAME)

    def list(self) -> List[Tab]:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            TabRepositoryInterface.list.__doc__
        )

        result_set = self.client.select(f"SELECT * FROM {TABLE_NAME}")

        list_: List[Tab] = []
        for row in result_set:
            list_.append(Tab(*row))

        return list_



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import logging
import sys
from collections import OrderedDict
from typing import Optional

from PyQt5.QtCore import QTimer
from PyQt5.QtGui import QFont
from PyQt5.QtWidgets import QApplication, QDialog, QLineEdit, QMessageBox, QWidget
from substrateinterface import Keypair, KeypairType

from tikka.domains.application import Application
from tikka.domains.entities.account import Account
from tikka.domains.entities.constants import DATA_PATH, WALLETS_PASSWORD_LENGTH
from tikka.libs.secret import generate_alphabetic, sanitize_mnemonic_string
from tikka.slots.pyqt.entities.constants import (
    ADDRESS_MONOSPACE_FONT_NAME,
    DEBOUNCE_TIME,
)
from tikka.slots.pyqt.resources.gui.windows.wallet_restore_rc import (
    Ui_WalletRestoreDialog,
)


class WalletRestoreWindow(QDialog, Ui_WalletRestoreDialog):
    """
    WalletRestoreWindow class
    """

    def __init__(
        self,
        application: Application,
        account: Account,
        reset_password: bool = False,
        parent: Optional[QWidget] = None,
    ):
        """
        Init import account window

        :param application: Application instance
        :param account: Account instance
        :param reset_password: Reset password if True (default to False)
        :param parent: QWidget instance
        """
        super().__init__(parent=parent)
        self.setupUi(self)

        self.application = application
        self.account = account
        self.reset_password = reset_password
        self._ = self.application.translator.gettext

        # set monospace font to address field
        monospace_font = QFont(ADDRESS_MONOSPACE_FONT_NAME)
        monospace_font.setStyleHint(QFont.Monospace)
        self.addressValueLabel.setFont(monospace_font)

        # Mnemonic language selector translated
        mnemonic_language_selector = OrderedDict(
            [
                ("en", self._("English")),
                ("fr", self._("French")),
                ("zh-hans", self._("Chinese simplified")),
                ("zh-hant", self._("Chinese traditional")),
                ("it", self._("Italian")),
                ("ja", self._("Japanese")),
                ("ko", self._("Korean")),
                ("es", self._("Spanish")),
            ]
        )
        for language_code, language_name in mnemonic_language_selector.items():
            self.mnemonicLanguageComboBox.addItem(language_name, userData=language_code)
        self.mnemonicLanguageComboBox.setCurrentIndex(
            self.mnemonicLanguageComboBox.findData(
                self.application.config.get("language")[:2]
            )
        )

        # buttons
        self.buttonBox.button(self.buttonBox.Ok).setEnabled(False)

        # events
        self.mnemonicLineEdit.textChanged.connect(self._on_mnemonic_changed)
        self.showButton.clicked.connect(self.on_show_button_clicked)
        self.mnemonicLanguageComboBox.currentIndexChanged.connect(
            self._on_mnemonic_changed
        )
        self.passwordChangeButton.clicked.connect(self._generate_wallet_password)
        self.buttonBox.accepted.connect(self.on_accepted_button)
        self.buttonBox.rejected.connect(self.close)

        # debounce timer
        self.mnemonic_debounce_timer = QTimer()
        self.mnemonic_debounce_timer.timeout.connect(self._mnemonic_debounce_call)

        # fill form
        self.storedPasswordFrame.hide()
        self.derivationValueLabel.setText(self.account.path)
        self.addressValueLabel.setText(self.account.address)
        self.nameValueLabel.setText(self.account.name)
        self._generate_wallet_password()

    def _on_mnemonic_changed(self):
        """
        Triggered when mnemonic is changed

        :return:
        """
        if self.mnemonic_debounce_timer.isActive():
            self.mnemonic_debounce_timer.stop()
        self.mnemonic_debounce_timer.start(DEBOUNCE_TIME)

    def _mnemonic_debounce_call(self):
        """
        Debounce function triggered only after idle time when typing mnemonic

        :return:
        """
        self._generate_address()

    def _generate_address(self) -> bool:
        """
        Generate address from mnemonic

        :return:
        """
        # stop debounce_timer to avoid infinite loop
        if self.mnemonic_debounce_timer.isActive():
            self.mnemonic_debounce_timer.stop()

        self.errorLabel.setText("")
        mnemonic = sanitize_mnemonic_string(self.mnemonicLineEdit.text())
        language_code = self.mnemonicLanguageComboBox.currentData()

        if not Keypair.validate_mnemonic(mnemonic, language_code):
            self.errorLabel.setText(self._("Mnemonic or language not valid!"))
            self.buttonBox.button(self.buttonBox.Ok).setEnabled(False)
            return False

        derivation_path = "" if self.account.path is None else self.account.path
        suri = mnemonic + derivation_path
        if suri == "":
            return False
        try:
            address = Keypair.create_from_uri(
                suri,
                ss58_format=self.application.currencies.get_current().ss58_format,
                crypto_type=KeypairType.SR25519,
                language_code=language_code,
            ).ss58_address
        except Exception as exception:
            logging.exception(exception)
            self.errorLabel.setText(self._("Mnemonic or language not valid!"))
            self.buttonBox.button(self.buttonBox.Ok).setEnabled(False)
            return False

        # if mnemonic address is not account address...
        if address != self.account.address:
            self.errorLabel.setText(
                self._("Mnemonic address is not the account address!")
            )
            self.buttonBox.button(self.buttonBox.Ok).setEnabled(False)
            return False

        # if password exists, hide the password field
        root_keypair = Keypair.create_from_mnemonic(
            mnemonic=mnemonic,
            language_code=language_code,
            crypto_type=KeypairType.SR25519,
            ss58_format=self.application.currencies.get_current().ss58_format,
        )
        if (
            self.application.passwords.exists(root_keypair.ss58_address)
            and self.reset_password is False
        ):
            stored_password = self.application.passwords.get_clear_password(
                root_keypair
            )
            self.storedpasswordLineEdit.setText(stored_password)
            self.storedPasswordFrame.show()
            self.passwordFrame.hide()
        else:
            self.storedPasswordFrame.hide()
            self.passwordFrame.show()

        self.buttonBox.button(self.buttonBox.Ok).setEnabled(True)
        return True

    def on_show_button_clicked(self):
        """
        Triggered when user click on show button

        :return:
        """
        if self.mnemonicLineEdit.echoMode() == QLineEdit.Password:
            self.mnemonicLineEdit.setEchoMode(QLineEdit.Normal)
            self.storedpasswordLineEdit.setEchoMode(QLineEdit.Normal)
            self.showButton.setText(self._("Hide"))
        else:
            self.mnemonicLineEdit.setEchoMode(QLineEdit.Password)
            self.storedpasswordLineEdit.setEchoMode(QLineEdit.Password)
            self.showButton.setText(self._("Show"))

    def _generate_wallet_password(self):
        """
        Generate new password for wallet encryption in UI

        :return:
        """
        self.passwordLineEdit.setText(generate_alphabetic(WALLETS_PASSWORD_LENGTH))

    def on_accepted_button(self):
        """
        Triggered when user click on ok button

        :return:
        """
        # user inputs
        mnemonic = sanitize_mnemonic_string(self.mnemonicLineEdit.text())
        derivation_path = "" if self.account.path is None else self.account.path
        suri = mnemonic + derivation_path
        language_code = self.mnemonicLanguageComboBox.currentData()

        # generated inputs
        password = self.passwordLineEdit.text()

        # root keypair for password
        root_keypair = Keypair.create_from_mnemonic(
            mnemonic=mnemonic,
            language_code=language_code,
            crypto_type=KeypairType.SR25519,
            ss58_format=self.application.currencies.get_current().ss58_format,
        )
        # if password exists for root account...
        if (
            self.application.passwords.exists(root_keypair.ss58_address)
            and self.reset_password is False
        ):
            # get stored password
            clear_password = self.application.passwords.get_clear_password(root_keypair)
            if clear_password is not None:
                password = clear_password
        else:
            # store new password
            self.application.passwords.new(root_keypair, password)

        # update all existing wallets for this root account and his derivation with new password
        if self.reset_password:
            self.application.accounts.reset_password(
                root_keypair, mnemonic, language_code, password
            )

        # create keypair from mnemonic to get seed as hexadecimal
        keypair = Keypair.create_from_uri(
            suri,
            language_code=language_code,
            crypto_type=KeypairType.SR25519,
            ss58_format=self.application.currencies.get_current().ss58_format,
        )

        wallet = self.application.wallets.get(keypair.ss58_address)
        if wallet is None:
            # create and store Wallet instance
            wallet = self.application.wallets.create(keypair, password)
            self.application.wallets.add(wallet)
        else:
            # display confirm dialog and get response
            button = QMessageBox.question(
                self,
                self._("Change wallet password?"),
                self._(
                    "A wallet already exists for this account. Change password for {address} wallet?"
                ).format(address=wallet.address),
            )
            if button == QMessageBox.Yes:
                # create and store Wallet instance
                new_wallet = self.application.wallets.create(keypair, password)
                self.application.wallets.update(new_wallet)

        self.application.accounts.unlock(self.account, password)


if __name__ == "__main__":
    qapp = QApplication(sys.argv)
    application_ = Application(DATA_PATH)
    account_ = Account(
        "5D2YjANfzuGvJWc3gykeFokJoboi57WAqbaVdo8VJzduFNsB", path="//2", name="test name"
    )
    WalletRestoreWindow(application_, account_).exec_()



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from typing import Optional

from substrateinterface import Keypair

from tikka.domains.entities.identity import STATUS_CREATED, STATUS_VALIDATED, Identity
from tikka.interfaces.adapters.network.identities import NetworkIdentitiesInterface
from tikka.interfaces.adapters.repository.identities import (
    IdentitiesRepositoryInterface,
)


class Identities:

    """
    Identities domain class
    """

    def __init__(
        self,
        repository: IdentitiesRepositoryInterface,
        network: NetworkIdentitiesInterface,
    ):
        """
        Init Accounts domain

        :param repository: AccountsRepositoryInterface instance
        :param network: NetworkAccountsInterface instance
        """
        self.repository = repository
        self.network = network

    @staticmethod
    def create(
        index: int,
        removable_on: int,
        next_creatable_on: int,
        status: str = STATUS_CREATED,
    ):
        """
        Return an identity instance from params

        :param index: Index number in blockchain
        :param removable_on: Identity expiration timestamp
        :param next_creatable_on: Date after which a new identity can be created
        :param status: Identity status
        :return:
        """
        return Identity(
            index=index,
            removable_on=removable_on,
            next_creatable_on=next_creatable_on,
            status=status,
        )

    def add(self, identity: Identity):
        """
        Add identity in repository

        :param identity: Identity instance
        :return:
        """
        self.repository.add(identity)

    def update(self, identity: Identity):
        """
        Update identity in repository

        :param identity: Identity instance
        :return:
        """
        self.repository.update(identity)

    def get(self, index: int) -> Optional[Identity]:
        """
        Get identity instance

        :param index: Identity index
        :return:
        """
        return self.repository.get(index)

    def delete(self, index: int) -> None:
        """
        Delete identity in repository

        :param index: Identity index to delete
        :return:
        """
        self.repository.delete(index)

    def exists(self, index: int) -> bool:
        """
        Return True if identity exists in repository

        :param index: Identity index to check
        :return:
        """
        return self.repository.exists(index)

    def is_validated(self, index: int) -> bool:
        """
        Return True if identity status is validated

        :param index: Identity index to check
        :return:
        """
        identity = self.get(index)
        if identity is None:
            return False
        return identity.status == STATUS_VALIDATED

    def fetch_index_from_network(self, address: str) -> Optional[int]:
        """
        Fetch account identity index from network if any

        :param address: Account address
        :return:
        """
        return self.network.get_identity_index(address)

    def fetch_from_network(self, index: int) -> Optional[Identity]:
        """
        Fetch Identity instance by index from network if any

        :param index: Identity index
        :return:
        """
        identity = self.network.get_identity(index)
        if identity is not None:
            if self.exists(index) is True:
                self.update(identity)
            else:
                self.add(identity)

        return identity

    def change_owner_key(self, old_keypair: Keypair, new_keypair: Keypair) -> bool:
        """
        Change identity owner from old_keypair to new_keypair

        :param old_keypair: Keypair of current identity account
        :param new_keypair: Keypair of new identity account
        :return:
        """
        return self.network.change_owner_key(old_keypair, new_keypair)



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
class DisplayAddress:
    """
    Class to handle display of wallet address
    """

    def __init__(self, address: str) -> None:
        """
        Init a DisplayAddress instance from the address

        :param address: Wallet address
        """
        self.address = address

    @property
    def shorten(self) -> str:
        """
        Return a shorten version of address (first 4 and last 4 characters)

        :return:
        """
        return f"{self.address[0:4]}…{self.address[-4:]}"

    def __str__(self) -> str:
        """
        Return string representation of instance

        :return:
        """
        return f"{self.address}"

    def __eq__(self, other) -> bool:
        """
        Test PublicKey equality

        :param other: Other PublicKey instance
        :return:
        """
        if not isinstance(other, DisplayAddress):
            return False

        if self.address == other.address:
            return True

        return False

    def __hash__(self):
        return self.address



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import random
from typing import Optional

from Crypto.Cipher import ChaCha20_Poly1305
from Crypto.Protocol.KDF import scrypt
from substrateinterface import Keypair

from tikka.domains.entities.constants import PASSWORDS_NONCE_SIZE
from tikka.domains.entities.password import Password
from tikka.interfaces.adapters.repository.passwords import PasswordsRepositoryInterface


class Passwords:
    """
    Passwords domain class
    """

    chacha_key_length = 32
    scrypt_n = 8
    scrypt_r = 1
    scrypt_p = 16

    def __init__(self, repository: PasswordsRepositoryInterface):
        """
        Init Passwords domain

        :param repository: Database adapter instance
        """
        self.repository = repository

    @staticmethod
    def create(keypair: Keypair, clear_password: str) -> Password:
        """
        Return an encrypted Password instance

        :param keypair: Keypair instance
        :param clear_password: Clear password
        :return:
        """
        nonce = random_bytes(PASSWORDS_NONCE_SIZE)
        key = scrypt(
            keypair.private_key.hex(),
            keypair.public_key.hex(),
            Passwords.chacha_key_length,
            Passwords.scrypt_n,
            Passwords.scrypt_r,
            Passwords.scrypt_p,
        )
        assert isinstance(key, bytes)
        cypher = ChaCha20_Poly1305.new(key=key, nonce=nonce)
        encrypted_password, mac_tag = cypher.encrypt_and_digest(
            clear_password.encode("utf-8")
        )
        return Password(
            root=keypair.ss58_address,
            encrypted_password=encrypted_password.hex(),
            encryption_nonce=nonce.hex(),
            encryption_mac_tag=mac_tag.hex(),
        )

    def new(self, keypair: Keypair, clear_password: str) -> Password:
        """
        Create Password instance and add it to repository
        if password already exists, update password

        :param keypair: Seed hexadecimal string
        :param clear_password: Password string
        :return:
        """
        password = self.create(keypair, clear_password)

        if self.exists(keypair.ss58_address):
            self.update(password)
        else:
            self.add(password)

        return password

    def add(self, password: Password):
        """
        Add Password instance in repository

        :param password: Password instance
        :return:
        """
        self.repository.add(password)

    def get(self, root: str) -> Optional[Password]:
        """
        Return Password instance from root address

        :param root: Root address
        :return:
        """
        return self.repository.get(root)

    def get_clear_password(self, keypair: Keypair) -> Optional[str]:
        """
        Get clear password from Keypair

        :param keypair: Keypair of root account
        :return:
        """
        password = self.get(keypair.ss58_address)
        if password is None:
            return None
        key = scrypt(
            keypair.private_key.hex(),
            keypair.public_key.hex(),
            Passwords.chacha_key_length,
            Passwords.scrypt_n,
            Passwords.scrypt_r,
            Passwords.scrypt_p,
        )
        assert isinstance(key, bytes)
        cypher = ChaCha20_Poly1305.new(
            key=key, nonce=bytes.fromhex(password.encryption_nonce)
        )
        return cypher.decrypt_and_verify(
            bytes.fromhex(password.encrypted_password),
            bytes.fromhex(password.encryption_mac_tag),
        ).decode("utf-8")

    def update(self, password: Password):
        """
        Update password

        :param password: Password instance
        :return:
        """
        self.repository.update(password)

    def delete(self, root: str) -> None:
        """
        Delete password in repository

        :param root: Root address of password to delete
        :return:
        """
        self.repository.delete(root)

    def exists(self, root: str) -> bool:
        """
        Return True if password with root address exists in repository

        :param root: Root address to check
        :return:
        """
        return self.repository.exists(root)


def random_bytes(size: int) -> bytes:
    """
    Return random bytes of given length

    :param size: Size of nonce in bytes
    :return:
    """
    return bytearray(random.getrandbits(8) for _ in range(size))



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import logging
import sys
from typing import Dict, Optional

from PyQt5.QtCore import QMutex
from PyQt5.QtGui import QFont, QMovie
from PyQt5.QtWidgets import QApplication, QCompleter, QDialog, QWidget
from substrateinterface import Keypair
from substrateinterface.utils.ss58 import is_valid_ss58_address

from tikka.domains.application import Application
from tikka.domains.entities.account import Account
from tikka.domains.entities.constants import AMOUNT_UNIT_KEY, DATA_PATH
from tikka.slots.pyqt.entities.constants import (
    ADDRESS_MONOSPACE_FONT_NAME,
    ICON_LOADER,
    SELECTED_UNIT_PREFERENCES_KEY,
    TRANSFER_RECIPIENT_ADDRESS_PREFERENCES_KEY,
    TRANSFER_SENDER_ADDRESS_PREFERENCES_KEY,
)
from tikka.slots.pyqt.entities.worker import AsyncQWorker
from tikka.slots.pyqt.resources.gui.windows.transfer_rc import Ui_TransferDialog
from tikka.slots.pyqt.windows.account_unlock import AccountUnlockWindow


class TransferWindow(QDialog, Ui_TransferDialog):
    """
    TransferWindow class
    """

    def __init__(
        self,
        application: Application,
        mutex: QMutex,
        sender: Optional[Account] = None,
        recipient: Optional[Account] = None,
        parent: Optional[QWidget] = None,
    ):
        """
        Init transfer window

        :param application: Application instance
        :param sender: Sender Account instance or None
        :param recipient: Recipient Account instance or None
        :param mutex: QMutex instance
        :param parent: QWidget instance
        """
        super().__init__(parent=parent)
        self.setupUi(self)

        self.application = application

        preferences_sender_address = self.application.preferences_repository.get(
            TRANSFER_SENDER_ADDRESS_PREFERENCES_KEY
        )
        if sender is None and preferences_sender_address is not None:
            preferences_sender_account = self.application.accounts.get_by_address(
                preferences_sender_address
            )
            if preferences_sender_account is not None:
                sender = preferences_sender_account

        preferences_recipient_address = self.application.preferences_repository.get(
            TRANSFER_RECIPIENT_ADDRESS_PREFERENCES_KEY
        )
        if recipient is None and preferences_recipient_address is not None:
            preferences_recipient_account = self.application.accounts.get_by_address(
                preferences_recipient_address
            )
            if preferences_recipient_account is not None:
                recipient = preferences_recipient_account

        self.sender_account = sender
        self.recipient_account = recipient
        self.mutex = mutex
        self._ = self.application.translator.gettext

        # substrate error message for gettext extraction
        self._("Transfer/payment would kill account")
        self._("Balance too low to send value")

        self.monospace_font = QFont(ADDRESS_MONOSPACE_FONT_NAME)
        self.monospace_font.setStyleHint(QFont.Monospace)

        # init unit from preference or default
        unit_preference = self.application.preferences_repository.get(
            SELECTED_UNIT_PREFERENCES_KEY
        )
        if unit_preference is not None:
            self.unit = unit_preference
        else:
            self.unit = AMOUNT_UNIT_KEY

        self.init_units()

        self.amount_value = 0
        self.fees = None
        self.transfer_success: Optional[bool] = None

        if self.sender_account is not None:
            if self.sender_account.name is None:
                self.senderNameOrAddressLineEdit.setText(self.sender_account.address)
                self.senderNameOrAddressLineEdit.setFont(self.monospace_font)
                self.senderNameOrAddressValueLabel.setText("")
            else:
                self.senderNameOrAddressLineEdit.setText(self.sender_account.name)
                self.senderNameOrAddressLineEdit.setFont(QFont())
                self.senderNameOrAddressValueLabel.setText(self.sender_account.address)
                self.senderNameOrAddressValueLabel.setFont(self.monospace_font)

        if self.recipient_account is not None:
            if self.recipient_account.name is None:
                self.recipientNameOrAddressLineEdit.setText(
                    self.recipient_account.address
                )
                self.recipientNameOrAddressLineEdit.setFont(self.monospace_font)
                self.recipientNameOrAddressValueLabel.setText("")
            else:
                self.recipientNameOrAddressLineEdit.setText(self.recipient_account.name)
                self.recipientNameOrAddressLineEdit.setFont(QFont())
                self.recipientNameOrAddressValueLabel.setText(
                    self.recipient_account.address
                )
                self.recipientNameOrAddressValueLabel.setFont(self.monospace_font)

        # animated loading icon
        self.loader_movie = QMovie(ICON_LOADER)
        self.loader_movie.start()
        self.loaderIconLabel.setMovie(self.loader_movie)
        loader_icon_size_policy = self.loaderIconLabel.sizePolicy()
        loader_icon_size_policy.setRetainSizeWhenHidden(True)
        self.loaderIconLabel.setSizePolicy(loader_icon_size_policy)
        self.loaderIconLabel.hide()

        # autocomplete  (dirty version)
        # todo: create a model to fetch only a filtered list from database on edit
        self.account_address_by_name: Dict[str, str] = {}
        account_address_by_name_duplicate_count = 1
        for account_instance in self.application.accounts.get_list():
            if (
                account_instance.name in self.account_address_by_name
                and account_instance.name is not None
            ):
                self.account_address_by_name[
                    f"{account_instance.name} ({account_address_by_name_duplicate_count})"
                ] = account_instance.address
            else:
                if account_instance.name is None:
                    self.account_address_by_name[""] = account_instance.address
                else:
                    self.account_address_by_name[
                        account_instance.name
                    ] = account_instance.address

        # sender autocomplete only on accounts with wallet
        self.sender_account_wordlist = []
        for account_name, account_address in self.account_address_by_name.items():
            if account_address in self.application.wallets.list_addresses():
                self.sender_account_wordlist.append(account_name)
                self.sender_account_wordlist.append(account_address)
        sender_completer = QCompleter(self.sender_account_wordlist)
        self.senderNameOrAddressLineEdit.setCompleter(sender_completer)

        # recipient autocomplete
        self.recipient_account_wordlist = list(
            self.account_address_by_name.keys()
        ) + list(self.account_address_by_name.values())
        recipient_completer = QCompleter(self.recipient_account_wordlist)
        self.recipientNameOrAddressLineEdit.setCompleter(recipient_completer)

        self._update_ui()

        # events
        self.senderNameOrAddressLineEdit.textChanged.connect(
            self._on_sender_address_line_edit_changed
        )
        self.recipientNameOrAddressLineEdit.textChanged.connect(
            self._on_recipient_address_line_edit_changed
        )
        self.amountDoubleSpinBox.valueChanged.connect(
            self._on_amount_double_spin_box_changed
        )
        self.amountUnitComboBox.activated.connect(self._on_unit_changed)
        self.feesButton.clicked.connect(self._on_fees_button_clicked)
        self.sendButton.clicked.connect(self._on_send_button_clicked)
        self.buttonBox.button(self.buttonBox.Close).clicked.connect(self.close)

        ##############################
        # ASYNC METHODS
        ##############################
        # fetch recipient balance
        self.fetch_recipient_balance_from_network_async_qworker = AsyncQWorker(
            self.fetch_recipient_balance_from_network, self.mutex, self
        )
        self.fetch_recipient_balance_from_network_async_qworker.finished.connect(
            self._on_finished_fetch_recipient_balance_from_network
        )
        # fetch sender balance
        self.fetch_sender_balance_from_network_async_qworker = AsyncQWorker(
            self.fetch_sender_balance_from_network, self.mutex, self
        )
        self.fetch_sender_balance_from_network_async_qworker.finished.connect(
            self._on_finished_fetch_sender_balance_from_network
        )

        # fetch fees from network
        self.fetch_fees_from_network_async_qworker = AsyncQWorker(
            self.fetch_fees_from_network, self.mutex, self
        )
        self.fetch_fees_from_network_async_qworker.finished.connect(
            self._on_finished_fetch_fees_from_network
        )
        # send transfer to network
        self.send_tranfer_to_network_async_qworker = AsyncQWorker(
            self.send_transfer_to_network, self.mutex, self
        )
        self.send_tranfer_to_network_async_qworker.finished.connect(
            self._on_finished_send_transfer_to_network
        )

    def _on_unit_changed(self):
        """
        Triggered when unit combo box is changed

        :return:
        """
        self.unit = self.amountUnitComboBox.currentData()
        self._update_ui()

    def _on_amount_double_spin_box_changed(self):
        """
        Triggered when the amount spin box is changed

        :return:
        """
        self.amount_value = self.amountDoubleSpinBox.value()
        self._update_ui()

    def fetch_recipient_balance_from_network(self):
        """
        Fetch last account data from the network

        :return:
        """
        self.loaderIconLabel.show()

        try:
            balance = self.application.accounts.fetch_balance_from_network(
                self.recipient_account.address
            )
        except Exception:
            # keep balance from cache DB if network failure
            balance = self.recipient_account.balance

        if balance is None:
            self.recipient_account.balance = 0
        else:
            self.recipient_account.balance = balance

    def _on_finished_fetch_recipient_balance_from_network(self):
        """
        Triggered when async request fetch_from_network is finished

        :return:
        """
        self.loaderIconLabel.hide()
        self._update_ui()

    def fetch_sender_balance_from_network(self):
        """
        Fetch last sender account data from the network

        :return:
        """
        self.loaderIconLabel.show()

        try:
            balance = self.application.accounts.fetch_balance_from_network(
                self.sender_account.address
            )
        except Exception:
            # keep balance from cache DB if network failure
            balance = self.sender_account.balance

        if balance is None:
            self.sender_account.balance = 0
        else:
            self.sender_account.balance = balance

    def _on_finished_fetch_sender_balance_from_network(self):
        """
        Triggered when async request fetch_from_network is finished

        :return:
        """
        self.loaderIconLabel.hide()
        self._update_ui()

    def _on_sender_address_line_edit_changed(self):
        """
        Triggered when text in the sender address field is changed

        :return:
        """
        self.errorLabel.setText("")
        address_or_name = self.senderNameOrAddressLineEdit.text().strip()

        try:
            address_is_valid = is_valid_ss58_address(address_or_name)
        except IndexError:
            address_is_valid = False

        if not address_is_valid:
            # if entry is a known account name...
            if address_or_name in self.account_address_by_name:
                self.senderNameOrAddressLineEdit.setFont(QFont())
                # display address under the name
                self.senderNameOrAddressValueLabel.setFont(self.monospace_font)
                self.senderNameOrAddressValueLabel.setText(
                    self.account_address_by_name[address_or_name]
                )
                self._get_sender_account(self.account_address_by_name[address_or_name])
            else:
                self.senderNameOrAddressValueLabel.setText("")
                self.senderBalanceValueLabel.setText("")
                self.errorLabel.setStyleSheet("color: red;")
                self.errorLabel.setText(
                    self._("Invalid address! Please check it again.")
                )
                self.sender_account = None

        else:
            self.senderNameOrAddressLineEdit.setFont(self.monospace_font)
            # display name under the address
            for key, value in self.account_address_by_name.items():
                if value == address_or_name:
                    self.senderNameOrAddressValueLabel.setFont(QFont())
                    self.senderNameOrAddressValueLabel.setText(key)
            self._get_sender_account(address_or_name)

        self._update_ui()

    def _on_recipient_address_line_edit_changed(self):
        """
        Triggered when text in the recipient address field is changed

        :return:
        """
        self.errorLabel.setText("")
        address_or_name = self.recipientNameOrAddressLineEdit.text().strip()

        try:
            address_is_valid = is_valid_ss58_address(address_or_name)
        except IndexError:
            address_is_valid = False

        if not address_is_valid:
            # if entry is a known account name...
            if address_or_name in self.account_address_by_name:
                self.recipientNameOrAddressLineEdit.setFont(QFont())
                # display address under the name
                self.recipientNameOrAddressValueLabel.setFont(self.monospace_font)
                self.recipientNameOrAddressValueLabel.setText(
                    self.account_address_by_name[address_or_name]
                )
                self._get_recipient_account(
                    self.account_address_by_name[address_or_name]
                )
            else:
                self.recipientNameOrAddressValueLabel.setText("")
                self.recipientBalanceValueLabel.setText("")
                self.errorLabel.setStyleSheet("color: red;")
                self.errorLabel.setText(
                    self._("Invalid address! Please check it again.")
                )
                self.recipient_account = None
        else:
            self.recipientNameOrAddressLineEdit.setFont(self.monospace_font)
            # display name under the address
            for key, value in self.account_address_by_name.items():
                if value == address_or_name:
                    self.recipientNameOrAddressValueLabel.setFont(QFont())
                    self.recipientNameOrAddressValueLabel.setText(key)
            self._get_recipient_account(address_or_name)

        self._update_ui()

    def _get_sender_account(self, address: str) -> None:
        """
        Get sender account

        :return:
        """
        self.sender_account = None
        try:
            Keypair(
                ss58_address=address,
                ss58_format=self.application.currencies.get_current().ss58_format,
            )
        except ValueError as exception:
            logging.exception(exception)
            self.transfer_success = False
            self.errorLabel.setStyleSheet("color: red;")
            self.errorLabel.setText(self._("Invalid address! Please check it again."))
            self.sender_account = None
            return None

        # save in preferences
        self.application.preferences_repository.set(
            TRANSFER_SENDER_ADDRESS_PREFERENCES_KEY, address
        )

        # search in local account list
        for account in self.application.accounts.get_list():
            if account.address == address:
                self.sender_account = account
                self._update_ui()
                return None

        # create account
        self.sender_account = Account(address)
        # fetch balance
        self.fetch_sender_balance_from_network_async_qworker.start()

        return None

    def _get_recipient_account(self, address: str) -> None:
        """
        Get recipient account

        :return:
        """
        self.recipient_account = None
        try:
            Keypair(
                ss58_address=address,
                ss58_format=self.application.currencies.get_current().ss58_format,
            )
        except ValueError as exception:
            logging.exception(exception)
            self.errorLabel.setStyleSheet("color: red;")
            self.errorLabel.setText(self._("Invalid address! Please check it again."))
            self.recipient_account = None
            return None

        # save in preferences
        self.application.preferences_repository.set(
            TRANSFER_RECIPIENT_ADDRESS_PREFERENCES_KEY, address
        )

        # search in local account list
        for account in self.application.accounts.get_list():
            if account.address == address:
                self.recipient_account = account
                self._update_ui()
                return None

        # create account
        self.recipient_account = Account(address)
        # fetch balance
        self.fetch_recipient_balance_from_network_async_qworker.start()

        return None

    def _update_ui(self):
        """
        Update UI

        :return:
        """
        # balances amount
        amount = self.application.amounts.get_amount(self.unit)
        if self.sender_account is not None:
            if self.sender_account.balance is None:
                self.senderBalanceValueLabel.setText("?")
            else:
                self.senderBalanceValueLabel.setText(
                    self.locale().toCurrencyString(
                        amount.value(self.sender_account.balance), amount.symbol()
                    )
                )

        if self.recipient_account is not None:
            if self.recipient_account.balance is None:
                self.recipientBalanceValueLabel.setText(self._("Unknown!"))
                self.recipientBalanceValueLabel.setToolTip(
                    self._(
                        "Account balance unknown! Send only one unit and make sure the owner can get it"
                    )
                )
            else:
                self.recipientBalanceValueLabel.setText(
                    self.locale().toCurrencyString(
                        amount.value(self.recipient_account.balance), amount.symbol()
                    )
                )
                self.recipientBalanceValueLabel.setToolTip("")

        # fees amount
        if self.fees is not None:
            self.feesValueLabel.setText(
                self.locale().toCurrencyString(amount.value(self.fees), amount.symbol())
            )
        else:
            self.feesValueLabel.setText("")

        # buttons
        if (
            self.sender_account is not None
            and self.recipient_account is not None
            and self.amount_value > 0
        ):
            self.sendButton.setDisabled(False)
            if self.application.wallets.is_unlocked(self.sender_account.address):
                self.feesButton.setDisabled(False)
        else:
            self.feesButton.setDisabled(True)
            self.sendButton.setDisabled(True)

    def init_units(self) -> None:
        """
        Init units combobox for transfer amount

        :return:
        """
        self.amountUnitComboBox.clear()

        for key, amount in self.application.amounts.register.items():
            self.amountUnitComboBox.addItem(amount.name(), userData=key)
        self.amountUnitComboBox.setCurrentIndex(
            self.amountUnitComboBox.findData(self.unit)
        )

    def _on_fees_button_clicked(self):
        """
        Triggered when user click on Fees button

        :return:
        """
        self.feesButton.setDisabled(True)
        self.loaderIconLabel.show()

        self.fetch_fees_from_network_async_qworker.start()

    def fetch_fees_from_network(self):
        """
        Fetch fees amount from network

        :return:
        """
        amount = self.application.amounts.get_amount(self.unit)
        # get value as blockchain units
        blockchain_value = amount.blockchain_value(self.amountDoubleSpinBox.value())
        sender_keypair = self.application.wallets.get_keypair(
            self.sender_account.address
        )
        self.fees = self.application.transfers.fees(
            sender_keypair, self.recipient_account.address, blockchain_value
        )

    def _on_finished_fetch_fees_from_network(self):
        """
        Triggered when async request fetch_from_network is finished

        :return:
        """
        self.loaderIconLabel.hide()
        self.feesButton.setDisabled(False)
        self._update_ui()

    def _on_send_button_clicked(self):
        """
        Triggered when user click on Send button

        :return:
        """
        # sender account locked...
        if not self.application.wallets.is_unlocked(self.sender_account.address):
            # ask password...
            dialog_code = AccountUnlockWindow(
                self.application, self.sender_account, self
            ).exec_()
            if dialog_code == QDialog.Rejected:
                return

        self.sendButton.setDisabled(True)
        self.loaderIconLabel.show()
        # send transfer to network
        self.send_tranfer_to_network_async_qworker.start()

    def send_transfer_to_network(self):
        """
        Send transfer to network

        :return:
        """
        self.errorLabel.setText("")
        # get value as blockchain units
        amount = self.application.amounts.get_amount(self.unit)
        blockchain_value = amount.blockchain_value(self.amountDoubleSpinBox.value())
        sender_keypair = self.application.wallets.get_keypair(
            self.sender_account.address
        )
        if sender_keypair is None:
            return
        extrinsic_receipt = self.application.transfers.send(
            sender_keypair, self.recipient_account.address, blockchain_value
        )
        if extrinsic_receipt is None:
            self.transfer_success = False
            self.errorLabel.setStyleSheet("color: red;")
            self.errorLabel.setText(
                self._("Transfer failed. Please check logs to understand why")
            )
        elif extrinsic_receipt.is_success is False:
            self.transfer_success = False
            self.errorLabel.setStyleSheet("color: red;")
            self.errorLabel.setText(self._(extrinsic_receipt.error_message["docs"][0]))
        else:
            self.transfer_success = True
            self.errorLabel.setStyleSheet("color: green;")
            self.errorLabel.setText(self._("Transfer done"))

    def _on_finished_send_transfer_to_network(self):
        """
        Triggered when async request send_transfer_to_network is finished

        :return:
        """
        self.loaderIconLabel.hide()
        self.sendButton.setDisabled(False)
        self.fetch_sender_balance_from_network_async_qworker.start()
        self.fetch_recipient_balance_from_network_async_qworker.start()


if __name__ == "__main__":
    qapp = QApplication(sys.argv)
    application_ = Application(DATA_PATH)
    account_ = application_.accounts.get_by_address(
        "5E6taVoL7eW8qpPKjDcf3GjyvFkkMB8EHkkqZksiWeViEkSP"
    )
    if account_ is not None:
        application_.accounts.delete(account_)
    application_.accounts.create_new_root_account(
        "album cute glance oppose hub fury strategy health dust rebuild trophy magic",
        "en",
        "test account",
        "aaaaaa",
    )
    account_ = application_.accounts.get_by_address(
        "5E6taVoL7eW8qpPKjDcf3GjyvFkkMB8EHkkqZksiWeViEkSP"
    )
    if account_ is not None:
        account_.balance = 0
        TransferWindow(application_, QMutex(), account_).exec_()



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import logging
from typing import Optional

from substrateinterface import Keypair
from substrateinterface.exceptions import SubstrateRequestException

from tikka.interfaces.adapters.network.authorities import NetworkAuthoritiesInterface


class NetworkAuthorities(NetworkAuthoritiesInterface):
    """
    NetworkAuthorities class
    """

    def rotate_keys(self) -> Optional[str]:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            NetworkAuthoritiesInterface.rotate_keys.__doc__
        )
        if not self.connections.is_connected():
            return None
        if self.connections.rpc.client is None:
            return None

        try:
            result = self.connections.rpc.client.rpc_request(
                "author_rotateKeys", []
            ).get("result")
        except Exception as exception:
            logging.exception(exception)
            return None

        return result

    def has_session_keys(self, session_keys: str) -> Optional[bool]:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            NetworkAuthoritiesInterface.has_session_keys.__doc__
        )
        if not self.connections.is_connected():
            return None
        if self.connections.rpc.client is None:
            return None

        try:
            result = self.connections.rpc.client.rpc_request(
                "author_hasSessionKeys", [session_keys]
            ).get("result")
        except Exception as exception:
            logging.exception(exception)
            return None

        return result

    def publish_session_keys(self, keypair: Keypair, session_keys: str) -> bool:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            NetworkAuthoritiesInterface.publish_session_keys.__doc__
        )
        if not self.connections.is_connected() or self.connections.rpc.client is None:
            return False

        session_keys_bytearray = bytearray.fromhex(session_keys[2:])
        params = {
            "keys": {
                "grandpa": f"0x{session_keys_bytearray[0:32].hex()}",
                "babe": f"0x{session_keys_bytearray[32:64].hex()}",
                "im_online": f"0x{session_keys_bytearray[64:96].hex()}",
                "authority_discovery": f"0x{session_keys_bytearray[96:128].hex()}",
            }
        }

        try:
            call = self.connections.rpc.client.compose_call(
                call_module="AuthorityMembers",
                call_function="set_session_keys",
                call_params=params,
            )
        except Exception as exception:
            logging.exception(exception)
            return False

        try:
            extrinsic = self.connections.rpc.client.create_signed_extrinsic(
                call=call, keypair=keypair
            )
        except Exception as exception:
            logging.exception(exception)
            return False

        try:
            # fixme: code stuck infinitely if no blocks are created on blockchain
            #       should have a timeout option
            result = self.connections.rpc.client.submit_extrinsic(
                extrinsic, wait_for_inclusion=True
            )
            logging.debug(
                "Extrinsic '%s' sent and included in block '%s'",
                result.extrinsic_hash,
                result.block_hash,
            )
        except SubstrateRequestException as exception:
            logging.exception(exception)
            return False

        if result.is_success is False:
            logging.error(result.error_message)

        return result.is_success

    def go_online(self, keypair: Keypair) -> bool:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            NetworkAuthoritiesInterface.go_online.__doc__
        )
        if not self.connections.is_connected() or self.connections.rpc.client is None:
            return False

        try:
            call = self.connections.rpc.client.compose_call(
                call_module="AuthorityMembers",
                call_function="go_online",
                call_params=None,
            )
        except Exception as exception:
            logging.exception(exception)
            return False

        try:
            extrinsic = self.connections.rpc.client.create_signed_extrinsic(
                call=call, keypair=keypair
            )
        except Exception as exception:
            logging.exception(exception)
            return False

        try:
            # fixme: code stuck infinitely if no blocks are created on blockchain
            #       should have a timeout option
            result = self.connections.rpc.client.submit_extrinsic(
                extrinsic, wait_for_inclusion=True
            )
            logging.debug(
                "Extrinsic '%s' sent and included in block '%s'",
                result.extrinsic_hash,
                result.block_hash,
            )
        except SubstrateRequestException as exception:
            logging.exception(exception)
            return False

        if result.is_success is False:
            logging.error(result.error_message)

        return result.is_success

    def go_offline(self, keypair: Keypair) -> bool:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            NetworkAuthoritiesInterface.go_offline.__doc__
        )
        if not self.connections.is_connected() or self.connections.rpc.client is None:
            return False

        try:
            call = self.connections.rpc.client.compose_call(
                call_module="AuthorityMembers",
                call_function="go_offline",
                call_params=None,
            )
        except Exception as exception:
            logging.exception(exception)
            return False

        try:
            extrinsic = self.connections.rpc.client.create_signed_extrinsic(
                call=call, keypair=keypair
            )
        except Exception as exception:
            logging.exception(exception)
            return False

        try:
            # fixme: code stuck infinitely if no blocks are created on blockchain
            #       should have a timeout option
            result = self.connections.rpc.client.submit_extrinsic(
                extrinsic, wait_for_inclusion=True
            )
            logging.debug(
                "Extrinsic '%s' sent and included in block '%s'",
                result.extrinsic_hash,
                result.block_hash,
            )
        except SubstrateRequestException as exception:
            logging.exception(exception)
            return False

        if result.is_success is False:
            logging.error(result.error_message)

        return result.is_success

    def is_online(self, identity_index: int) -> Optional[bool]:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            NetworkAuthoritiesInterface.is_online.__doc__
        )
        if not self.connections.is_connected():
            return None
        if self.connections.rpc.client is None:
            return None

        try:
            result = self.connections.rpc.client.query(
                "AuthorityMembers", "OnlineAuthorities"
            )
        except Exception as exception:
            logging.exception(exception)
            return None

        return identity_index in result.value

    def is_incoming(self, identity_index: int) -> Optional[bool]:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            NetworkAuthoritiesInterface.is_incoming.__doc__
        )
        if not self.connections.is_connected():
            return None
        if self.connections.rpc.client is None:
            return None

        try:
            result = self.connections.rpc.client.query(
                "AuthorityMembers", "IncomingAuthorities"
            )
        except Exception as exception:
            logging.exception(exception)
            return None

        return identity_index in result.value

    def is_outgoing(self, identity_index: int) -> Optional[bool]:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            NetworkAuthoritiesInterface.is_outgoing.__doc__
        )
        if not self.connections.is_connected():
            return None
        if self.connections.rpc.client is None:
            return None

        try:
            result = self.connections.rpc.client.query(
                "AuthorityMembers", "OutgoingAuthorities"
            )
        except Exception as exception:
            logging.exception(exception)
            return None

        return identity_index in result.value



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import logging
from typing import Optional

from tikka.domains.entities.node import Node
from tikka.interfaces.adapters.network.nodes import NetworkNodesInterface


class NetworkNodes(NetworkNodesInterface):
    """
    NetworkNodes class
    """

    def get(self) -> Optional[Node]:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            NetworkNodesInterface.get.__doc__
        )
        if not self.connections.is_connected():
            return None
        if self.connections.rpc.client is None:
            return None

        try:
            peer_id = self.connections.rpc.client.rpc_request(
                "system_localPeerId", []
            ).get("result")
        except Exception as exception:
            logging.exception(exception)
            return None

        try:
            current_block = self.connections.rpc.client.rpc_request(
                "system_syncState", []
            ).get("result")["currentBlock"]
        except Exception as exception:
            logging.exception(exception)
            return None

        return Node(
            self.connections.rpc.client.url,
            peer_id=peer_id,
            block=current_block,
            software=self.connections.rpc.client.name,
            software_version=self.connections.rpc.client.version,
        )



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import logging
import struct
from typing import Optional

from substrateinterface import Keypair
from substrateinterface.exceptions import SubstrateRequestException

from tikka.domains.entities.identity import Identity
from tikka.interfaces.adapters.network.identities import NetworkIdentitiesInterface


class NetworkIdentities(NetworkIdentitiesInterface):
    """
    NetworkIdentities class
    """

    def get_identity_index(self, address: str) -> Optional[int]:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            NetworkIdentitiesInterface.get_identity_index.__doc__
        )
        if not self.connections.is_connected():
            return None
        if self.connections.rpc.client is None:
            return None

        try:
            result = self.connections.rpc.client.query(
                "Identity", "IdentityIndexOf", [address]
            )
        except Exception as exception:
            logging.exception(exception)
            return None

        return result.value

    def get_identity(self, index: int) -> Optional[Identity]:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            NetworkIdentitiesInterface.get_identity.__doc__
        )
        if not self.connections.is_connected():
            return None
        if self.connections.rpc.client is None:
            return None

        try:
            result = self.connections.rpc.client.query(
                "Identity", "Identities", [index]
            )
        except Exception as exception:
            logging.exception(exception)
            return None
        return Identity(
            index=index,
            next_creatable_on=result["next_creatable_identity_on"].value,
            removable_on=int(result["removable_on"].value),
            status=result["status"].value,
        )

    def change_owner_key(self, old_keypair: Keypair, new_keypair: Keypair) -> bool:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            NetworkIdentitiesInterface.change_owner_key.__doc__
        )
        if not self.connections.is_connected():
            return False
        if self.connections.rpc.client is None:
            return False

        identity_index = self.get_identity_index(old_keypair.ss58_address)
        if identity_index is None:
            return False

        # message to sign
        prefix_bytes = b"icok"
        genesis_hash_str = self.connections.rpc.client.get_block_hash(0)
        genesis_hash_bytes = bytearray.fromhex(genesis_hash_str[2:])
        identity_index_bytes = struct.pack("<I", identity_index)
        identity_pubkey_bytes = old_keypair.public_key
        message_bytes = (
            prefix_bytes
            + genesis_hash_bytes
            + identity_index_bytes
            + identity_pubkey_bytes
        )

        # message signed by the new owner
        signature_bytes = new_keypair.sign(message_bytes)

        # newKey: AccountId32, newKeySig: SpRuntimeMultiSignature
        params = {
            "new_key": new_keypair.ss58_address,
            "new_key_sig": {"Sr25519": signature_bytes},
        }
        try:
            # create raw call (extrinsic)
            call = self.connections.rpc.client.compose_call(
                call_module="Identity",
                call_function="change_owner_key",
                call_params=params,
            )
        except Exception as exception:
            logging.exception(exception)
            return False

        try:
            # create extrinsic signed by current owner
            extrinsic = self.connections.rpc.client.create_signed_extrinsic(
                call=call, keypair=old_keypair
            )
        except Exception as exception:
            logging.exception(exception)
            return False

        try:
            response = self.connections.rpc.client.submit_extrinsic(
                extrinsic, wait_for_inclusion=True
            )
        except SubstrateRequestException as exception:
            logging.exception(exception)
            return False

        if response.is_success is False:
            logging.error(response.error_message)

        return response.is_success



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import abc
from typing import Optional

from substrateinterface import Keypair

from tikka.domains.entities.identity import Identity
from tikka.interfaces.domains.connections import ConnectionsInterface


class NetworkIdentitiesInterface(abc.ABC):
    """
    NetworkIdentitiesInterface class
    """

    def __init__(self, connections: ConnectionsInterface) -> None:
        """
        Use connections to request identities information

        :param connections: ConnectionsInterface instance
        :return:
        """
        self.connections = connections

    @abc.abstractmethod
    def get_identity_index(self, address: str) -> Optional[int]:
        """
        Return the account Identity instance from address if exists

        :param address: Account address
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_identity(self, index: int) -> Optional[Identity]:
        """
        Return the account Identity instance from index if exists

        :param index: Identity index
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def change_owner_key(self, old_keypair: Keypair, new_keypair: Keypair) -> bool:
        """
        Change identity owner from old_keypair to new_keypair

        :param old_keypair: Keypair of current identity account
        :param new_keypair: Keypair of new identity account
        :return:
        """
        raise NotImplementedError



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Optional

from tikka.adapters.repository.sqlite3 import Sqlite3RepositoryInterface
from tikka.domains.entities.identity import Identity
from tikka.interfaces.adapters.repository.identities import (
    IdentitiesRepositoryInterface,
)

TABLE_NAME = "identities"


class Sqlite3IdentitiesRepository(
    IdentitiesRepositoryInterface, Sqlite3RepositoryInterface
):
    """
    Sqlite3IdentitiesRepository class
    """

    def add(self, identity: Identity) -> None:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            IdentitiesRepositoryInterface.add.__doc__
        )

        # insert only non hidden fields
        self.client.insert(
            TABLE_NAME,
            **get_fields_from_identity(identity),
        )

    def get(self, index: int) -> Optional[Identity]:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            IdentitiesRepositoryInterface.get.__doc__
        )

        row = self.client.select_one(
            f"SELECT * FROM {TABLE_NAME} WHERE index_=?", (index,)
        )
        if row is None:
            return None

        return Identity(*row)

    def update(self, identity: Identity) -> None:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            IdentitiesRepositoryInterface.update.__doc__
        )

        # update only non hidden fields
        self.client.update(
            TABLE_NAME,
            f"index_='{identity.index}'",
            **get_fields_from_identity(identity),
        )

    def delete(self, index: int) -> None:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            IdentitiesRepositoryInterface.delete.__doc__
        )

        self.client.delete(TABLE_NAME, index_=index)

    def exists(self, index: int) -> bool:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            IdentitiesRepositoryInterface.exists.__doc__
        )

        row = self.client.select_one(
            f"SELECT count(index_) FROM {TABLE_NAME} WHERE index_=?", (index,)
        )
        if row is None:
            return False

        return row[0] == 1


def get_fields_from_identity(identity: Identity) -> dict:
    """
    Return a dict of supported fields with normalized value

    :param identity: Identity instance
    :return:
    """
    fields = {}
    for (key, value) in identity.__dict__.items():
        if key.startswith("_"):
            continue
        if key == "index":
            # index is a reserved keyword in sqlite3
            key = "index_"
        fields[key] = value

    return fields



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from dataclasses import dataclass
from typing import Optional, TypeVar

SmithType = TypeVar("SmithType", bound="Smith")


@dataclass
class SmithMembership:
    expire_on: int


@dataclass
class Smith:
    identity_index: int
    membership: Optional[SmithMembership]
    pending_membership: bool


@dataclass
class SmithCertification:
    issuer_identity_index: int
    issuer_address: str
    receiver_identity_index: int
    receiver_address: str
    expire_on_block: int



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from typing import Optional

from substrateinterface import ExtrinsicReceipt, Keypair

from tikka.domains.entities.events import TransferEvent
from tikka.domains.events import EventDispatcher
from tikka.domains.wallets import Wallets
from tikka.interfaces.adapters.network.transfers import NetworkTransfersInterface


class Transfers:
    """
    Transfers domain class
    """

    def __init__(
        self,
        wallets: Wallets,
        network: NetworkTransfersInterface,
        event_dispatcher: EventDispatcher,
    ):
        """
        Init Transfers domain

        :param wallets: Wallets domain
        :param network: NetworkTransfersInterface adapter instance
        :param event_dispatcher: EventDispatcher instance
        """
        self.wallets = wallets
        self.network = network
        self.event_dispatcher = event_dispatcher

    def fees(
        self, sender_keypair: Keypair, recipient_address: str, amount: int
    ) -> Optional[int]:
        """
        Fetch transfer fees from network and return it if request is successful

        :param sender_keypair: Sender Keypair instance
        :param recipient_address: Recipient address
        :param amount: Amount in blockchain unit
        :return:
        """
        return self.network.fees(sender_keypair, recipient_address, amount)

    def send(
        self, sender_keypair: Keypair, recipient_address: str, amount: int
    ) -> Optional[ExtrinsicReceipt]:
        """
        Send transfer to network and return ExtrinsicReceipt if request is successful

        :param sender_keypair: Sender Keypair instance
        :param recipient_address: Recipient address
        :param amount: Amount in blockchain unit
        :return:
        """
        receipt = self.network.send(sender_keypair, recipient_address, amount)

        if receipt is not None and receipt.is_success is True:
            # dispatch event
            event = TransferEvent(
                TransferEvent.EVENT_TYPE_SENT,
            )
            self.event_dispatcher.dispatch_event(event)

        return receipt

    def send_all(
        self, sender_keypair: Keypair, recipient_address: str, keep_alive: bool = False
    ) -> Optional[ExtrinsicReceipt]:
        """
        Send transfer to network and return ExtrinsicReceipt if request is successful

        :param sender_keypair: Sender Keypair instance
        :param recipient_address: Recipient address
        :param keep_alive: Optional, default False
        :return:
        """
        receipt = self.network.transfer_all(
            sender_keypair, recipient_address, keep_alive
        )

        if receipt is not None and receipt.is_success is True:
            # dispatch event
            event = TransferEvent(
                TransferEvent.EVENT_TYPE_SENT,
            )
            self.event_dispatcher.dispatch_event(event)

        return receipt



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Fonts
ADDRESS_MONOSPACE_FONT_NAME = "Courier"

# icon paths
ICON_ACCOUNT_NO_WALLET = ":/images/locked_no_wallet"
ICON_ACCOUNT_WALLET_LOCKED = ":/images/locked"
ICON_ACCOUNT_WALLET_UNLOCKED = ":/images/unlocked"
ICON_NETWORK_CONNECTED = ":/icons/connected"
ICON_NETWORK_DISCONNECTED = ":/icons/disconnected"
ICON_IDENTITY = ":/icons/member_icon"
ICON_LOADER = ":/icons/loader"

# no activity delay before executing function when typing in ms
DEBOUNCE_TIME = 300

# preferences keys
SELECTED_TAB_PAGE_PREFERENCES_KEY = "selected_tab_page"
WALLET_LOAD_DEFAULT_DIRECTORY_PREFERENCES_KEY = "wallet_load_default_directory"
WALLET_SAVE_DEFAULT_DIRECTORY_PREFERENCES_KEY = "wallet_save_default_directory"
SELECTED_UNIT_PREFERENCES_KEY = "selected_unit"
ACCOUNTS_TABLE_SORT_COLUMN_PREFERENCES_KEY = "accounts_table_sort_column"
ACCOUNTS_TABLE_SORT_ORDER_PREFERENCES_KEY = "accounts_table_sort_order"
ACCOUNTS_TABLE_CATEGORY_FILTER_PREFERENCES_KEY = "accounts_table_category_filter"
ACCOUNTS_TABLE_WALLET_FILTER_PREFERENCES_KEY = "accounts_table_wallet_filter"
TRANSFER_SENDER_ADDRESS_PREFERENCES_KEY = "transfer_sender_address"
TRANSFER_RECIPIENT_ADDRESS_PREFERENCES_KEY = "transfer_recipient_address"
SMITH_SELECTED_ACCOUNT_ADDRESS = "smith_selected_account_address"



#!/usr/bin/env python3

# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging

from tikka.domains.application import Application
from tikka.domains.entities.constants import DATA_PATH
from tikka.slots.pyqt.main import PyQtMainApplication

logging.basicConfig(level=logging.DEBUG)


def main():
    """
    Main function for build executable

    [tool.poetry.scripts]
    tikka = "tikka.__main__:main"

    :return:
    """
    # create domain application
    application_ = Application(DATA_PATH)
    # init main application
    PyQtMainApplication(application_)


if __name__ == "__main__":
    main()



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import logging
import sys
from typing import List, Optional

from PyQt5.QtCore import QLocale, QTimer
from PyQt5.QtGui import QMovie
from PyQt5.QtWidgets import (
    QApplication,
    QDialog,
    QMainWindow,
    QTableWidgetItem,
    QWidget,
)

from tikka.domains.application import Application
from tikka.domains.entities.account import Account
from tikka.domains.entities.constants import DATA_PATH
from tikka.domains.entities.events import ConnectionsEvent, CurrencyEvent
from tikka.domains.entities.node import Node
from tikka.domains.entities.smith import SmithCertification, SmithMembership
from tikka.slots.pyqt.entities.constants import (
    ICON_LOADER,
    SMITH_SELECTED_ACCOUNT_ADDRESS,
)
from tikka.slots.pyqt.resources.gui.widgets.smith_rc import Ui_SmithWidget
from tikka.slots.pyqt.windows.account_unlock import AccountUnlockWindow


class SmithWidget(QWidget, Ui_SmithWidget):
    """
    SmithWidget class
    """

    DELAY_BEFORE_UPDATE_MEMBERSHIP_STATUS_AFTER_REQUEST = 6000

    AUTHORITY_STATUS_NO = 0
    AUTHORITY_STATUS_INCOMING = 1
    AUTHORITY_STATUS_ONLINE = 2
    AUTHORITY_STATUS_OUTGOING = 3

    def __init__(
        self, application: Application, parent: Optional[QWidget] = None
    ) -> None:
        """
        Init SmithWidget instance

        :param application: Application instance
        :param parent: MainWindow instance
        """
        super().__init__(parent=parent)
        self.setupUi(self)

        self.application = application
        self._ = self.application.translator.gettext
        self.account: Optional[Account] = None
        self.membership: Optional[SmithMembership] = None
        self.pending_membership: bool = False
        self.authority_status: int = 0
        self.certs_by_receiver: Optional[List[SmithCertification]] = None
        self.current_node = self.application.nodes.get(
            self.application.nodes.get_current_url()
        )

        # animated loading icon
        self.loader_movie = QMovie(ICON_LOADER)
        self.loader_movie.start()
        self.loaderIconLabel.setMovie(self.loader_movie)
        loader_icon_size_policy = self.loaderIconLabel.sizePolicy()
        loader_icon_size_policy.setRetainSizeWhenHidden(True)
        self.loaderIconLabel.setSizePolicy(loader_icon_size_policy)
        self.loaderIconLabel.hide()

        # setup certifier table widget
        self.certifiersTableWidget.setColumnCount(3)
        self.certifiersTableWidget.setHorizontalHeaderLabels(
            [self._("Identity"), self._("Account"), self._("Expire on")]
        )

        ##############################
        # ASYNC METHODS
        ##############################
        self.fetch_membership_from_network_timer = QTimer()
        self.fetch_membership_from_network_timer.timeout.connect(
            self.fetch_membership_from_network_timer_function
        )
        self.fetch_authority_status_from_network_timer = QTimer()
        self.fetch_authority_status_from_network_timer.timeout.connect(
            self.fetch_authority_status_from_network_timer_function
        )
        self.rotate_keys_timer = QTimer()
        self.rotate_keys_timer.timeout.connect(self.rotate_keys_timer_function)
        self.publish_keys_timer = QTimer()
        self.publish_keys_timer.timeout.connect(self.publish_keys_timer_function)
        self.request_membership_timer = QTimer()
        self.request_membership_timer.timeout.connect(
            self.request_membership_timer_function
        )
        self.claim_membership_timer = QTimer()
        self.claim_membership_timer.timeout.connect(
            self.claim_membership_timer_function
        )
        self.revoke_membership_timer = QTimer()
        self.revoke_membership_timer.timeout.connect(
            self.revoke_membership_timer_function
        )
        self.go_online_timer = QTimer()
        self.go_online_timer.timeout.connect(self.go_online_timer_function)
        self.go_offline_timer = QTimer()
        self.go_offline_timer.timeout.connect(self.go_offline_timer_function)

        # events
        self.accountComboBox.activated.connect(self.on_account_combobox_index_changed)
        self.refreshSmithButton.clicked.connect(self._on_refresh_smith_button_clicked)
        self.rotateKeysButton.clicked.connect(self.rotate_keys_timer.start)
        self.publishKeysButton.clicked.connect(self.publish_keys_timer.start)
        self.requestMembershipButton.clicked.connect(
            self.request_membership_timer.start
        )
        self.revokeSmithMembershipButton.clicked.connect(
            self.revoke_membership_timer.start
        )
        self.claimMembershipButton.clicked.connect(self.claim_membership_timer.start)
        self.goOnlineButton.clicked.connect(self.go_online_timer.start)
        self.goOfflineButton.clicked.connect(self.go_offline_timer.start)

        # subscribe to application events
        self.application.event_dispatcher.add_event_listener(
            CurrencyEvent.EVENT_TYPE_CHANGED, self._on_currency_event
        )
        self.application.event_dispatcher.add_event_listener(
            ConnectionsEvent.EVENT_TYPE_CONNECTED, self.on_connections_event
        )
        self.application.event_dispatcher.add_event_listener(
            ConnectionsEvent.EVENT_TYPE_DISCONNECTED, self.on_connections_event
        )

        # populate form
        self.init_account_combo_box()
        if self.account is not None:
            self.fetch_membership_from_network_timer.start()
            self.fetch_authority_status_from_network_timer.start()
        else:
            self._update_ui()

    def init_account_combo_box(self) -> None:
        """
        Init combobox with validated identity accounts (with wallets)

        :return:
        """
        self.accountComboBox.clear()
        self.accountComboBox.addItem("-", userData=None)

        accounts = self.application.accounts.get_list()
        for account in accounts:
            if (
                account.identity_index is not None
                and self.application.identities.is_validated(account.identity_index)
                and self.application.wallets.exists(account.address)
            ):
                self.accountComboBox.addItem(
                    account.name if account.name is not None else account.address,
                    userData=account.address,
                )

        preference_account_address_selected = self.application.preferences.get(
            SMITH_SELECTED_ACCOUNT_ADDRESS
        )
        if preference_account_address_selected is not None:
            preference_account_selected = self.application.accounts.get_by_address(
                preference_account_address_selected
            )
            if preference_account_selected is not None:
                self.accountComboBox.setCurrentIndex(
                    accounts.index(preference_account_selected)
                )
                self.account = preference_account_selected

    def on_account_combobox_index_changed(self):
        """
        Triggered when account selection is changed

        :return:
        """
        address = self.accountComboBox.currentData()
        if address is not None:
            self.account = self.application.accounts.get_by_address(address)
        else:
            self.account = None
            self.membership = None
            self.pending_membership = False
            self.certs_by_receiver = None

        self._update_ui()

        if self.account is not None:
            self.loaderIconLabel.show()
            # Disable button
            self.refreshSmithButton.setEnabled(False)
            self.fetch_membership_from_network_timer.start()
            self.fetch_authority_status_from_network_timer.start()

    def rotate_keys_timer_function(self):
        """
        Triggered when user click on rotate keys button

        :return:
        """
        self.rotateKeysButton.setDisabled(True)
        result = self.application.authorities.rotate_keys(self.current_node)
        if result is not None:
            self.sessionKeysTextBrowser.setText(result)
            self.errorLabel.setText("")
        else:
            self.errorLabel.setText(self._("Unable to get session keys"))

        self.rotateKeysButton.setEnabled(True)
        self.rotate_keys_timer.stop()

    def request_membership_timer_function(self):
        """
        Triggered when user click on request membership button

        :return:
        """
        if self.account is None:
            return
        # if account is locked...
        if not self.application.wallets.is_unlocked(self.account.address):
            # ask password...
            dialog_code = AccountUnlockWindow(
                self.application, self.account, self
            ).exec_()
            if dialog_code == QDialog.Rejected:
                return

        self.requestMembershipButton.setDisabled(True)
        result = self.application.smiths.request_membership(
            self.application.wallets.get_keypair(self.account.address),
            self.current_node.session_keys,
        )
        if not result:
            self.errorLabel.setText(self._("Unable to request membership"))

        self.requestMembershipButton.setEnabled(True)
        self.fetch_membership_from_network_timer.start(
            self.DELAY_BEFORE_UPDATE_MEMBERSHIP_STATUS_AFTER_REQUEST
        )

        self.request_membership_timer.stop()

    def revoke_membership_timer_function(self):
        """
        Triggered when user click on revoke membership button

        :return:
        """
        if self.account is None:
            return
        # if account is locked...
        if not self.application.wallets.is_unlocked(self.account.address):
            # ask password...
            dialog_code = AccountUnlockWindow(
                self.application, self.account, self
            ).exec_()
            if dialog_code == QDialog.Rejected:
                return

        self.revokeSmithMembershipButton.setDisabled(True)
        result = self.application.smiths.revoke_membership(
            self.application.wallets.get_keypair(self.account.address),
        )
        if not result:
            self.errorLabel.setText(self._("Unable to revoke smith membership"))

        self.revokeSmithMembershipButton.setEnabled(True)
        self.fetch_membership_from_network_timer.start(
            self.DELAY_BEFORE_UPDATE_MEMBERSHIP_STATUS_AFTER_REQUEST
        )

        self.revoke_membership_timer.stop()

    def claim_membership_timer_function(self):
        """
        Triggered when user click on claim membership button

        :return:
        """
        if self.account is None:
            return
        # if account is locked...
        if not self.application.wallets.is_unlocked(self.account.address):
            # ask password...
            dialog_code = AccountUnlockWindow(
                self.application, self.account, self
            ).exec_()
            if dialog_code == QDialog.Rejected:
                return

        self.claimMembershipButton.setDisabled(True)
        result = self.application.smiths.claim_membership(
            self.application.wallets.get_keypair(self.account.address)
        )
        if not result:
            self.errorLabel.setText(self._("Unable to claim membership"))

        self.claimMembershipButton.setEnabled(True)
        self.fetch_membership_from_network_timer.start(
            self.DELAY_BEFORE_UPDATE_MEMBERSHIP_STATUS_AFTER_REQUEST
        )

        self.claim_membership_timer.stop()

    def publish_keys_timer_function(self):
        """
        Triggered when user click on publish keys button

        :return:
        """
        if self.account is None:
            return
        # if account is locked...
        if not self.application.wallets.is_unlocked(self.account.address):
            # ask password...
            dialog_code = AccountUnlockWindow(
                self.application, self.account, self
            ).exec_()
            if dialog_code == QDialog.Rejected:
                return

        self.publishKeysButton.setDisabled(True)
        result = self.application.authorities.publish_session_keys(
            self.application.wallets.get_keypair(self.account.address),
            self.current_node.session_keys,
        )
        if not result:
            self.errorLabel.setText(self._("Unable to publish session keys"))

        self.publishKeysButton.setEnabled(True)
        self.publish_keys_timer.stop()

    def go_online_timer_function(self):
        """
        Triggered when user click on go online button

        :return:
        """
        if self.account is None:
            return
        # if account is locked...
        if not self.application.wallets.is_unlocked(self.account.address):
            # ask password...
            dialog_code = AccountUnlockWindow(
                self.application, self.account, self
            ).exec_()
            if dialog_code == QDialog.Rejected:
                return

        self.goOnlineButton.setDisabled(True)
        result = self.application.authorities.go_online(
            self.application.wallets.get_keypair(self.account.address)
        )
        if not result:
            self.errorLabel.setText(self._("Unable to claim membership"))

        self.goOnlineButton.setEnabled(True)
        self.fetch_membership_from_network_timer.start(
            self.DELAY_BEFORE_UPDATE_MEMBERSHIP_STATUS_AFTER_REQUEST
        )

        self.go_online_timer.stop()

    def go_offline_timer_function(self):
        """
        Triggered when user click on go offline button

        :return:
        """
        if self.account is None:
            return
        # if account is locked...
        if not self.application.wallets.is_unlocked(self.account.address):
            # ask password...
            dialog_code = AccountUnlockWindow(
                self.application, self.account, self
            ).exec_()
            if dialog_code == QDialog.Rejected:
                return

        self.goOfflineButton.setDisabled(True)
        result = self.application.authorities.go_offline(
            self.application.wallets.get_keypair(self.account.address)
        )
        if not result:
            self.errorLabel.setText(self._("Unable to claim membership"))

        self.goOfflineButton.setEnabled(True)
        self.fetch_membership_from_network_timer.start(
            self.DELAY_BEFORE_UPDATE_MEMBERSHIP_STATUS_AFTER_REQUEST
        )

        self.go_offline_timer.stop()

    def _on_refresh_smith_button_clicked(self, _):
        """ """
        self.loader_movie.start()
        self.loaderIconLabel.show()
        self.fetch_membership_from_network_timer.start()
        self.fetch_authority_status_from_network_timer.start()

    def fetch_membership_from_network_timer_function(self):
        """
        Update smith membership infos from current url connection

        :return:
        """
        if self.account is None or self.account.identity_index is None:
            return

        # Start the thread
        # get membership status
        self.membership = self.application.smiths.fetch_membership_from_network(
            self.account.identity_index
        )
        if self.membership is None:
            # get pending status
            self.pending_membership = (
                self.application.smiths.fetch_pending_membership_from_network(
                    self.account.identity_index
                )
            )

        if self.pending_membership is True or self.membership is not None:
            # get certification list
            self.certs_by_receiver = (
                self.application.smiths.fetch_certs_by_receiver_from_network(
                    self.account.address, self.account.identity_index
                )
            )

        self.refreshSmithButton.setEnabled(True)
        self.loaderIconLabel.hide()

        self._update_ui()

        self.fetch_membership_from_network_timer.stop()

    def fetch_authority_status_from_network_timer_function(self):
        """
        Update authority status for account identity from current url connection

        :return:
        """
        if self.account is None or self.account.identity_index is None:
            return

        # Start the thread
        self.authority_status = self.AUTHORITY_STATUS_NO

        is_online = self.application.authorities.fetch_is_online_from_network(
            self.account.identity_index
        )
        if is_online is True:
            self.authority_status = self.AUTHORITY_STATUS_ONLINE
        else:
            is_incoming = self.application.authorities.fetch_is_incoming_from_network(
                self.account.identity_index
            )
            if is_incoming is True:
                self.authority_status = self.AUTHORITY_STATUS_INCOMING
            else:
                is_outgoing = (
                    self.application.authorities.fetch_is_outgoing_from_network(
                        self.account.identity_index
                    )
                )
                if is_outgoing is True:
                    self.authority_status = self.AUTHORITY_STATUS_OUTGOING

        self.loaderIconLabel.hide()

        self._update_ui()

        self.fetch_authority_status_from_network_timer.stop()

    def _update_ui(self):
        """
        Update node infos in UI

        :return:
        """
        self.errorLabel.setText("")
        self.urlValueLabel.setText(self.current_node.url)

        if self.current_node.session_keys is not None:
            self.sessionKeysTextBrowser.setText(self.current_node.session_keys)
        else:
            self.sessionKeysTextBrowser.setText("")

        if not self.application.connections.is_connected():
            self.rotateKeysButton.setDisabled(True)
            self.refreshSmithButton.setDisabled(True)
            self.publishKeysButton.setDisabled(True)
            self.requestMembershipButton.setDisabled(True)
            self.claimMembershipButton.setDisabled(True)
            self.revokeSmithMembershipButton.setDisabled(True)
            self.goOnlineButton.setDisabled(True)
            self.goOfflineButton.setDisabled(True)
            self.errorLabel.setText(self._("No network connection"))
            return

        # rotate keys only available on localhost via an ssh tunnel or other method...
        self.rotateKeysButton.setEnabled(
            "localhost" in self.application.nodes.get_current_url()
        )

        # disable all buttons
        self.refreshSmithButton.setDisabled(True)
        self.publishKeysButton.setDisabled(True)
        self.requestMembershipButton.setDisabled(True)
        self.claimMembershipButton.setDisabled(True)
        self.revokeSmithMembershipButton.setDisabled(True)
        self.goOnlineButton.setDisabled(True)
        self.goOfflineButton.setDisabled(True)

        if self.account is not None:
            self.refreshSmithButton.setEnabled(True)
            if self.current_node.session_keys is not None:
                if self.membership is not None:
                    self.publishKeysButton.setEnabled(True)
            if self.membership is None and self.pending_membership is False:
                self.requestMembershipButton.setEnabled(True)
            if self.membership is None and self.pending_membership is True:
                self.claimMembershipButton.setEnabled(True)
            if self.membership is not None:
                self.revokeSmithMembershipButton.setEnabled(True)
                self.goOnlineButton.setEnabled(
                    self.authority_status == self.AUTHORITY_STATUS_NO
                )
                self.goOfflineButton.setEnabled(
                    self.authority_status == self.AUTHORITY_STATUS_ONLINE
                )

        if self.membership is not None:
            # display membership expiration block number
            # QLocale.toString(app_utils.get_default_locale(), d, 'dd MMM')
            datetime_from_block = self.application.block_time.get_datetime_from_block(
                self.membership.expire_on
            )
            if datetime_from_block is not None:
                expire_on_localized_datetime_string = self.locale().toString(
                    datetime_from_block,
                    QLocale.dateFormat(self.locale(), QLocale.ShortFormat),
                )
                self.membershipValueLabel.setText(
                    self._("expire on {datetime}").format(
                        datetime=expire_on_localized_datetime_string
                    )
                )
            else:
                self.membershipValueLabel.setText(
                    self._("expire on block #{block}").format(
                        block=self.membership.expire_on
                    )
                )
            self.membershipValueLabel.setToolTip(
                self._("block #{block}").format(block=self.membership.expire_on)
            )
        elif self.pending_membership is True:
            self.membershipValueLabel.setText(self._("Pending..."))
        else:
            self.membershipValueLabel.setText(self._("No"))

        if self.authority_status == self.AUTHORITY_STATUS_NO:
            self.authorityValueLabel.setText(self._("No"))
        elif self.authority_status == self.AUTHORITY_STATUS_INCOMING:
            self.authorityValueLabel.setText(self._("Incoming..."))
        elif self.authority_status == self.AUTHORITY_STATUS_ONLINE:
            self.authorityValueLabel.setText(self._("Online"))
        elif self.authority_status == self.AUTHORITY_STATUS_OUTGOING:
            self.authorityValueLabel.setText(self._("Outgoing..."))

        # clear certification table
        self.certifiersTableWidget.clearContents()

        if self.certs_by_receiver is not None:
            self.certifiersTableWidget.setRowCount(len(self.certs_by_receiver))
            for row_index, smith_certification in enumerate(self.certs_by_receiver):
                identity_item = QTableWidgetItem(
                    str(smith_certification.issuer_identity_index)
                )
                address_item = QTableWidgetItem(str(smith_certification.issuer_address))

                datetime_from_block = (
                    self.application.block_time.get_datetime_from_block(
                        smith_certification.expire_on_block
                    )
                )
                if datetime_from_block is not None:
                    expire_on_localized_datetime_string = self.locale().toString(
                        datetime_from_block,
                        QLocale.dateFormat(self.locale(), QLocale.ShortFormat),
                    )
                    expire_on_item = QTableWidgetItem(
                        expire_on_localized_datetime_string
                    )
                    expire_on_item.setToolTip(
                        self._("block #{block}").format(
                            block=smith_certification.expire_on_block
                        )
                    )
                else:
                    expire_on_item = QTableWidgetItem(
                        f"#{smith_certification.expire_on_block}"
                    )

                self.certifiersTableWidget.setItem(row_index, 0, identity_item)
                self.certifiersTableWidget.setItem(row_index, 1, address_item)
                self.certifiersTableWidget.setItem(row_index, 2, expire_on_item)

            self.certifiersTableWidget.resizeColumnsToContents()
            self.certifiersTableWidget.horizontalHeader().stretchLastSection()

    def _on_currency_event(self, _):
        """
        When a currency event is triggered

        :param _: CurrencyEvent instance
        :return:
        """
        self.init_account_combo_box()
        if self.account is not None:
            self.fetch_membership_from_network_timer.start()
        else:
            self._update_ui()

    def on_connections_event(self, _):
        """
        Triggered when the network connection if connected/disconnected

        :param _: ConnectionsEvent instance
        :return:
        """

        self.current_node = self.application.nodes.get(
            self.application.nodes.get_current_url()
        )
        self.init_account_combo_box()
        self._update_ui()


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    qapp = QApplication(sys.argv)

    application_ = Application(DATA_PATH)

    main_window = QMainWindow()
    main_window.show()
    validator_node = Node(url="ws://localhost:9944")
    if application_.nodes.get(validator_node.url) is None:
        application_.nodes.add(validator_node)
    application_.nodes.set_current_url(validator_node.url)
    main_window.setCentralWidget(SmithWidget(application_, main_window))

    sys.exit(qapp.exec_())



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import abc
from typing import Optional

from substrateinterface import Keypair

from tikka.interfaces.domains.connections import ConnectionsInterface


class NetworkAuthoritiesInterface(abc.ABC):
    """
    NetworkAuthoritiesInterface class
    """

    def __init__(self, connections: ConnectionsInterface) -> None:
        """
        Use connections to request/send online authorities information

        :param connections: ConnectionsInterface instance
        :return:
        """
        self.connections = connections

    @abc.abstractmethod
    def rotate_keys(self) -> Optional[str]:
        """
        Rotate Session keys

        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def has_session_keys(self, session_keys: str) -> Optional[bool]:
        """
        Return True if the current node keystore store private session keys corresponding to public session_keys

        :param session_keys: Session public keys (hex string "0x123XYZ")
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def publish_session_keys(self, keypair: Keypair, session_keys: str) -> bool:
        """
        Set/Change in blockchain the session public keys for the Keypair account

        :param keypair: Owner Keypair
        :param session_keys: Session public keys (hex string "0x123XYZ")
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def go_online(self, keypair: Keypair) -> bool:
        """
        Start writing blocks with smith account from keypair

        :param keypair: Smith account Keypair
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def go_offline(self, keypair: Keypair) -> bool:
        """
        Stop writing blocks with smith account from keypair

        :param keypair: Smith account Keypair
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def is_online(self, identity_index: int) -> Optional[bool]:
        """
        Return True if identity_index is in online authorities

        :param identity_index: Identity index
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def is_incoming(self, identity_index: int) -> Optional[bool]:
        """
        Return True if identity_index is in incoming authorities

        :param identity_index: Identity index
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def is_outgoing(self, identity_index: int) -> Optional[bool]:
        """
        Return True if identity_index is in outgoing authorities

        :param identity_index: Identity index
        :return:
        """
        raise NotImplementedError



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import abc
from typing import List, Optional

from substrateinterface import Keypair

from tikka.domains.entities.smith import SmithCertification, SmithMembership
from tikka.interfaces.domains.connections import ConnectionsInterface


class NetworkSmithsInterface(abc.ABC):
    """
    NetworkSmithsInterface class
    """

    def __init__(self, connections: ConnectionsInterface) -> None:
        """
        Use connections to request/send smiths information

        :param connections: ConnectionsInterface instance
        :return:
        """
        self.connections = connections

    @abc.abstractmethod
    def request_membership(self, keypair: Keypair, session_keys: str) -> bool:
        """
        Request a smith membership for the Keypair account with node session_keys

        :param keypair: Owner Keypair
        :param session_keys: Session public keys (hex string "0x123XYZ")
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def claim_membership(self, keypair: Keypair) -> bool:
        """
        Claim that last smith membership request for the Keypair account fulfill all requirements

        :param keypair: Owner Keypair
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def revoke_membership(self, keypair: Keypair) -> bool:
        """
        Revoke smith membership for the Keypair account

        :param keypair: Owner Keypair
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def membership(self, identity_index: int) -> Optional[SmithMembership]:
        """
        Return SmithMembership instance

        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def pending_membership(self, identity_index: int) -> bool:
        """
        Return True if identity status is pending_membership

        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def certs_by_receiver(
        self, receiver_address: str, receiver_identity_index: int
    ) -> Optional[List[SmithCertification]]:
        """
        Return a list of certification received by identity_index

         [
         [identity index, expire on block number],
         [identity index, expire on block number]
         ]

        :param receiver_address: Address of account receiving certs
        :param receiver_identity_index: Index of identity receiving certs
        :return:
        """
        raise NotImplementedError



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import logging
import sys
from collections import OrderedDict
from typing import Optional

from PyQt5.QtCore import Qt, QTimer
from PyQt5.QtGui import QFont, QKeyEvent
from PyQt5.QtWidgets import QApplication, QDialog, QLineEdit, QWidget
from substrateinterface import Keypair, KeypairType

from tikka.domains.application import Application
from tikka.domains.entities.account import Account
from tikka.domains.entities.constants import DATA_PATH
from tikka.libs.secret import sanitize_mnemonic_string
from tikka.slots.pyqt.entities.constants import (
    ADDRESS_MONOSPACE_FONT_NAME,
    DEBOUNCE_TIME,
)
from tikka.slots.pyqt.resources.gui.windows.account_derivation_create_rc import (
    Ui_AccountDerivationCreateDialog,
)


class AccountDerivationCreateWindow(QDialog, Ui_AccountDerivationCreateDialog):
    """
    AccountDerivationCreateWindow class
    """

    def __init__(
        self,
        application: Application,
        account: Account,
        parent: Optional[QWidget] = None,
    ):
        """
        Init import account window

        :param application: Application instance
        :param account: Account instance
        :param parent: QWidget instance
        """
        super().__init__(parent=parent)
        self.setupUi(self)

        self.application = application
        self.account = account
        self._ = self.application.translator.gettext
        self.keypair = None
        self.password = None

        # Mnemonic language selector translated
        mnemonic_language_selector = OrderedDict(
            [
                ("en", self._("English")),
                ("fr", self._("French")),
                ("zh-hans", self._("Chinese simplified")),
                ("zh-hant", self._("Chinese traditional")),
                ("it", self._("Italian")),
                ("ja", self._("Japanese")),
                ("ko", self._("Korean")),
                ("es", self._("Spanish")),
            ]
        )
        for language_code, language_name in mnemonic_language_selector.items():
            self.mnemonicLanguageComboBox.addItem(language_name, userData=language_code)
        self.mnemonicLanguageComboBox.setCurrentIndex(
            self.mnemonicLanguageComboBox.findData(
                self.application.config.get("language")[:2]
            )
        )

        # derivation selector
        for derivation_path in self.application.accounts.get_available_derivation_list(
            self.account
        ):
            self.derivationComboBox.addItem(derivation_path, userData=derivation_path)
        monospace_font = QFont(ADDRESS_MONOSPACE_FONT_NAME)
        monospace_font.setStyleHint(QFont.Monospace)
        self.addressValueLabel.setFont(monospace_font)
        self.rootAddressValueLabel.setFont(monospace_font)

        # buttons
        self.buttonBox.button(self.buttonBox.Ok).setEnabled(False)

        # events
        self.mnemonicLineEdit.textChanged.connect(self._on_mnemonic_changed)
        self.showButton.clicked.connect(self.on_show_button_clicked)
        self.mnemonicLanguageComboBox.currentIndexChanged.connect(
            self._on_mnemonic_changed
        )
        self.derivationComboBox.lineEdit().keyPressEvent = (
            self._on_derivation_keypress_event
        )
        self.derivationComboBox.currentIndexChanged.connect(self._generate_address)
        self.buttonBox.accepted.connect(self.on_accepted_button)
        self.buttonBox.rejected.connect(self.close)

        # debounce timer
        self.mnemonic_debounce_timer = QTimer()
        self.mnemonic_debounce_timer.timeout.connect(self._mnemonic_debounce_call)

        # fill form
        self.rootAddressValueLabel.setText(self.account.address)

    def _on_derivation_keypress_event(self, event: QKeyEvent):
        """

        :param event:
        :return:
        """
        if event.key() == Qt.Key_Return:
            return
        # if the key is not return, handle normally
        QLineEdit.keyPressEvent(self.derivationComboBox.lineEdit(), event)
        self._generate_address()

    def _on_mnemonic_changed(self):
        """
        Triggered when mnemonic is changed

        :return:
        """
        if self.mnemonic_debounce_timer.isActive():
            self.mnemonic_debounce_timer.stop()
        self.mnemonic_debounce_timer.start(DEBOUNCE_TIME)

    def _mnemonic_debounce_call(self):
        """
        Debounce function triggered only after idle time when typing mnemonic

        :return:
        """
        self._generate_address()

    def _generate_address(self) -> bool:
        """
        Generate address from mnemonic

        :return:
        """
        # stop mnemonic_debounce_timer to avoid infinite loop
        if self.mnemonic_debounce_timer.isActive():
            self.mnemonic_debounce_timer.stop()

        self.errorLabel.setText("")
        language_code = self.mnemonicLanguageComboBox.currentData()
        mnemonic = sanitize_mnemonic_string(self.mnemonicLineEdit.text())
        derivation_path = self.derivationComboBox.currentText()
        suri = mnemonic + derivation_path
        if not Keypair.validate_mnemonic(mnemonic, language_code):
            self.errorLabel.setText(self._("Mnemonic or language not valid!"))
            self.buttonBox.button(self.buttonBox.Ok).setEnabled(False)
            return False

        try:
            root_keypair = Keypair.create_from_mnemonic(
                mnemonic=mnemonic,
                ss58_format=self.application.currencies.get_current().ss58_format,
                crypto_type=KeypairType.SR25519,
                language_code=language_code,
            )
        except Exception as exception:
            logging.exception(exception)
            self.errorLabel.setText(self._("Mnemonic or language not valid!"))
            self.buttonBox.button(self.buttonBox.Ok).setEnabled(False)
            return False

        # if mnemonic address is not account address...
        if root_keypair.ss58_address != self.account.address:
            self.errorLabel.setText(
                self._("Mnemonic address is not the root account address!")
            )
            self.buttonBox.button(self.buttonBox.Ok).setEnabled(False)
            return False

        try:
            self.keypair = Keypair.create_from_uri(
                suri=suri,
                ss58_format=self.application.currencies.get_current().ss58_format,
                crypto_type=KeypairType.SR25519,
                language_code=language_code,
            )
        except Exception as exception:
            logging.exception(exception)
            self.errorLabel.setText(self._("Mnemonic or language not valid!"))
            self.buttonBox.button(self.buttonBox.Ok).setEnabled(False)
            return False

        if self.keypair is not None:
            self.addressValueLabel.setText(self.keypair.ss58_address)

            if (
                self.application.accounts.get_by_address(self.keypair.ss58_address)
                is not None
            ):
                self.errorLabel.setText(self._("Account already exists!"))
                self.buttonBox.button(self.buttonBox.Ok).setEnabled(False)
                return False

            stored_password = self.application.passwords.get_clear_password(
                root_keypair
            )
            self.storedpasswordLineEdit.setText(stored_password)
            self.buttonBox.button(self.buttonBox.Ok).setEnabled(True)

        return True

    def on_show_button_clicked(self):
        """
        Triggered when user click on show button

        :return:
        """
        if self.mnemonicLineEdit.echoMode() == QLineEdit.Password:
            self.mnemonicLineEdit.setEchoMode(QLineEdit.Normal)
            self.storedpasswordLineEdit.setEchoMode(QLineEdit.Normal)
            self.showButton.setText(self._("Hide"))
        else:
            self.mnemonicLineEdit.setEchoMode(QLineEdit.Password)
            self.storedpasswordLineEdit.setEchoMode(QLineEdit.Password)
            self.showButton.setText(self._("Show"))

    def on_accepted_button(self):
        """
        Triggered when user click on ok button

        :return:
        """
        # user inputs
        mnemonic = sanitize_mnemonic_string(self.mnemonicLineEdit.text())
        derivation_path = self.derivationComboBox.currentText()
        language_code = self.mnemonicLanguageComboBox.currentData()
        name = self.nameLineEdit.text().strip()

        # generated inputs
        root_keypair = Keypair.create_from_mnemonic(
            mnemonic=mnemonic,
            language_code=language_code,
            crypto_type=KeypairType.SR25519,
            ss58_format=self.application.currencies.get_current().ss58_format,
        )
        password = self.application.passwords.get_clear_password(root_keypair)
        assert password is not None

        # create derived account + wallet (and the read-only root account if not exists)
        self.application.accounts.create_new_account(
            mnemonic, language_code, derivation_path, name, password
        )


if __name__ == "__main__":
    qapp = QApplication(sys.argv)
    application_ = Application(DATA_PATH)
    account_ = Account(
        "5E6taVoL7eW8qpPKjDcf3GjyvFkkMB8EHkkqZksiWeViEkSP", name="test name"
    )
    AccountDerivationCreateWindow(application_, account_).exec_()



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import logging
import sys
from typing import List, Optional

from PyQt5.QtGui import QFont
from PyQt5.QtWidgets import QApplication, QDialog, QWidget
from substrateinterface.utils.ss58 import is_valid_ss58_address

from tikka.domains.application import Application
from tikka.domains.entities.account import Account
from tikka.domains.entities.constants import DATA_PATH
from tikka.domains.entities.events import AccountEvent
from tikka.slots.pyqt.entities.constants import ADDRESS_MONOSPACE_FONT_NAME
from tikka.slots.pyqt.resources.gui.windows.scan_qrcode_rc import Ui_ScanQRCodeDialog


class ScanQRCodeOpenCVWindow(QDialog, Ui_ScanQRCodeDialog):
    """
    ScanQRCodeOpenCVWindow class
    """

    def __init__(self, application: Application, parent: Optional[QWidget] = None):
        """
        Init scan qrcode OpenCV window

        :param application: Application instance
        :param parent: QWidget instance
        """
        super().__init__(parent=parent)
        self.setupUi(self)

        self.application = application
        self._ = self.application.translator.gettext
        self.address: Optional[str] = None

        # set monospace font to address field
        monospace_font = QFont(ADDRESS_MONOSPACE_FONT_NAME)
        monospace_font.setStyleHint(QFont.Monospace)
        self.addressValueLabel.setFont(monospace_font)

        # buttons
        self.buttonBox.button(self.buttonBox.Ok).setEnabled(False)

        # events
        self.buttonBox.accepted.connect(self.on_accepted_button)
        self.buttonBox.rejected.connect(self.close)

        camera_index = self.get_available_camera_opencv_index()
        if camera_index is not None:
            self.address = self.opencv_webcam_qrcode_scanner(camera_index)
            if self.address is not None:
                self.addressValueLabel.setText(self.address)
                self.errorLabel.setText("")
                self.buttonBox.button(self.buttonBox.Ok).setEnabled(True)
            else:
                self.errorLabel.setText(self._("QRCode Scanner canceled"))
        else:
            self.errorLabel.setText(self._("No camera available"))

    @staticmethod
    def opencv_webcam_qrcode_scanner(camera_index: int) -> Optional[str]:
        """
        Open OpenCV webcam window to scan QR Code

        :param camera_index: Camera index to use as scanner
        :return:
        """
        # if OpenCV lib is imported at start of the module, this error occurs :
        #
        # QObject::moveToThread: Current thread (0x55a5fbae7cc0) is not the object's thread (0x55a5fc1c79f0).
        # Cannot move to target thread (0x55a5fbae7cc0)
        #
        # qt.qpa.plugin: Could not load the Qt platform plugin "xcb" in
        # "/home/vit/Documents/dev/python/tikka/.venv/lib/python3.7/site-packages/cv2/qt/plugins"
        # even though it was found.
        # This application failed to start because no Qt platform plugin could be initialized. Reinstalling the
        # application may fix this problem.
        #
        # Available platform plugins are: xcb, eglfs, linuxfb, minimal, minimalegl, offscreen, vnc, wayland-egl,
        # wayland, wayland-xcomposite-egl, wayland-xcomposite-glx, webgl.
        #
        # So we import OpenCV lib here, as a dirty trick to display OpenCV window in Qt GUI
        import cv2

        address: Optional[str] = None
        detector = cv2.QRCodeDetector()
        cap = cv2.VideoCapture(camera_index)
        while True:
            _, img = cap.read()
            # detect and decode
            data, bbox, _ = detector.detectAndDecode(img)
            # check if there is a QRCode in the image
            if data:
                if is_valid_ss58_address(data):
                    address = data
                    break
            cv2.imshow("QRCODE Scanner", img)
            key = cv2.waitKey(1)
            # print(key)
            # close window to stop loop
            if cv2.getWindowProperty("QRCODE Scanner", cv2.WND_PROP_VISIBLE) < 1:
                break
            # Esc key to stop loop
            if key == 27:
                break

        cap.release()
        cv2.destroyAllWindows()

        return address

    @staticmethod
    def get_available_camera_opencv_index() -> Optional[int]:
        """
        Return first available working camera index or None

        :return:
        """
        import cv2

        non_working_ports: List[int] = []
        dev_port = 0
        working_ports: List[int] = []
        available_ports: List[int] = []
        while (
            len(non_working_ports) < 5
        ):  # if there are more than 5 non working ports stop the testing.
            camera = cv2.VideoCapture(dev_port)
            if not camera.isOpened():
                non_working_ports.append(dev_port)
                logging.debug("Port %s is not working." % dev_port)
            else:
                is_reading, img = camera.read()
                width = camera.get(3)
                height = camera.get(4)
                if is_reading:
                    logging.debug(
                        "Camera %s is working and reads images (%s x %s)"
                        % (dev_port, width, height)
                    )
                    working_ports.append(dev_port)
                else:
                    logging.debug(
                        "Port %s for camera ( %s x %s) is present but does not reads."
                        % (dev_port, width, height)
                    )
                    available_ports.append(dev_port)
            dev_port += 1
        if len(working_ports) > 0:
            return working_ports[0]
        return None

    def on_accepted_button(self) -> None:
        """
        Triggered when user click on ok button

        :return:
        """
        if self.address is not None:
            account = self.application.accounts.get_by_address(self.address)
            if account is None:
                # create account instance
                account = Account(self.address, name=self._("Added from a QRCode"))

                # add instance in application
                self.application.accounts.add(account)
            else:
                # dispatch event
                event = AccountEvent(
                    AccountEvent.EVENT_TYPE_ADD,
                    account,
                )
                self.application.event_dispatcher.dispatch_event(event)


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    qapp = QApplication(sys.argv)
    application_ = Application(DATA_PATH)
    window = ScanQRCodeOpenCVWindow(application_)
    if window.address is not None:
        # display window
        window.exec_()
